import numpy as np
import matplotlib.pyplot as plt

from lib5c.util.plotting import plotter


@plotter
def plot_fdr(eval_results, labels, colors=None, p_alt=None, **kwargs):
    """
    Plots an FDR control curve from a list of ``eval.npz``-style results.

    Parameters
    ----------
    eval_results : list of dict-like
        The dicts should have keys 'thresh' and 'fdr' whose values are parallel
        vectors describing the thresholds and FDRs to use for the FDR control
        curve. Each dict in the list represents a different FDR control curve
        which will be overlayed in the plot.
    labels : list of str
        List of labels parallel to ``eval_results`` providing names for each
        curve.
    colors : list of valid matplotlib colors, optional
        Colors for each FDR curve. Pass None to automatically assign colors.
    p_alt : float, optional
        Pass the true proportion of alternative (non-null) points to draw a
        dashed line representing the optimal BH-FDR control line and shade the
        zone of successful FDR control. Pass None to draw a dashed line
        indicating the boundary of successful FDR control.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis
        The axis plotted on.
    """
    # resolve colors
    if colors is None:
        colors = ['C%i' % i for i in range(len(labels))]

    for i, (res, label, color) in enumerate(zip(eval_results, labels, colors)):
        # unbox results
        fdr = res['fdr']
        thresh = res['thresh']

        # plot
        fdr_idx = np.isfinite(fdr)
        plt.plot(1 - thresh[fdr_idx], fdr[fdr_idx], color=color, label=label)

    # dashed line and shading (if requested)
    if p_alt is None:
        plt.plot([0, 1], [0, 1], color='gray', linestyle='--', zorder=1)
    else:
        plt.fill_between([0, 1], [0, 0], [0, 1], color='lightgray')
        plt.plot([0, 1], [0, 1-p_alt], color='gray', linestyle='--', zorder=1)

    # limits, labels, legend
    plt.axis('scaled')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.xlabel('FDR threshold')
    plt.ylabel('FDR')
    plt.legend(loc='upper left')



import numpy as np
import matplotlib.pyplot as plt

from lib5c.util.plotting import plotter
from lib5c.plotters.scatter import scatter

from hic3defdr.util.scaled_nb import mvr, inverse_mvr


@plotter
def compare_disp_fits(fit_fns, labels, max_dist=200, colors=None,
                      linestyles=None, legend=True, **kwargs):
    """
    Compares multiple dispersion fit functions.

    Parameters
    ----------
    fit_fns : list of functions
        The fit functions to compare. Each function should be vectorized and
        take one argument (the distance) and return one value (the fitted
        dispersion at that distance).
    labels : list of str
        The labels to use for each fit function being compared.
    max_dist : int
        The maximum distance out to which the plot should be drawn.
    colors : list of valid matplotlib colors, optional
        Pass a list of colors to color each fit function with. Pass None to
        color them automatically.
    linestyles : list of valid matplotlib linestyles, optional
        Pass a list of linestyles to plot each fit function with. Pass None to
        use the default linestyle.
    legend : bool
        Pass True to include a legend on the plot.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis
        The axis plotted on.
    """
    # resolve colors
    if colors is None:
        colors = ['C%s' % i for i in range(len(fit_fns))]

    # resolve linestyles
    if linestyles is None:
        linestyles = [None] * len(fit_fns)

    # define x values
    xs = np.arange(max_dist + 1)

    # plot
    for f, l, c, ls in zip(fit_fns, labels, colors, linestyles):
        plt.plot(xs, f(xs), color=c, linestyle=ls, label=l)

    # labels
    plt.xlabel('distance')
    plt.ylabel('dispersion')

    # add legend
    if legend:
        plt.legend()


@plotter
def plot_mvr(pixel_mean, pixel_var=None, pixel_disp=None, pixel_dist=None,
             pixel_var_fit=None, pixel_disp_fit=None, mean_per_bin=None,
             dist_per_bin=None, var_per_bin=None, disp_per_bin=None,
             fit_align_dist=False, xaxis='mean', yaxis='var', dist_max=200,
             mean_min=5.0, scatter_fit=-1, scatter_size=36,
             hexbin=True, logx=True, logy=True, xlim=None, ylim=None,
             legend=True, **kwargs):
    """
    Plots pixel-wise, bin-wise, and estimated dispersions in terms of either
    dispersion or variance versus either pixel-wise mean or distance.

    Parameters
    ----------
    pixel_mean : np.ndarray
        The pixel-wise mean.
    pixel_var, pixel_disp : np.ndarray, optional
        The pixel-wise variance and dispersion. Pass only one of these.
    pixel_dist : np.ndarray, optional
        The pixel-wise distance. Not needed for simple MVR plotting.
    pixel_var_fit, pixel_disp_fit : np.ndarray, optional
        The smoothed pixel-wise variance or dispersion estimates. Pass only one
        of these.
    mean_per_bin, dist_per_bin : np.ndarray, optional
        The mean or distance of each bin used for estimating the dispersions but
        before any smoothing was performed. Pass only one of these.
    var_per_bin, disp_per_bin : np.ndarray, optional
        The estimated variance or dispersion of each bin before any smoothing
        was performed. Pass only one of these.
    fit_align_dist : bool
        Pass True if the var/disp was fitted as a function of distance, in which
        case the fitted vars/disps will be aligned to distance rather than in a
        pixel-wise fashion. Use this to fix "jagged" fit lines caused by
        setting ``xaxis`` to a different value than the one the fitting was
        actually performed against.
    xaxis : 'mean' or 'dist'
        What to plot on the x-axis.
    yaxis : 'var' or 'disp'
        What to plot on the y-axis.
    dist_max : int
        If ``xaxis`` is "dist", the maximum distance to plot in bin units.
    mean_min : float
        If `xaxis` is "mean", the minimum mean to plot.
    scatter_fit : int
        Pass a nonzero integer to draw the fitted vars/disps as a scatter plot
        of ``scatter_fit`` selected points. Pass -1 to plot the fitted
        vars/disps as a curve. Pass 0 to omit plotting the var/disp estimates
        altogether.
    scatter_size : int
        The marker size when plotting scatter plots.
    hexbin : bool
        Pass False to skip plotting the hexbins, leaving only the estimated
        variances or dispersions.
    logx, logy : bool
        Whether or not to log the x- or y-axis, respectively.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis
        The axis plotted on.
    """
    # convert pixel_* to [unique_]dist_* quantities part 1: dist_mean (dd)
    # dist_* quantities are aligned to np.arange(pixel_dist.max() + 1)
    # dist_* quantities are only made if pixel_dist is passed
    # it's really nice if dist_mean is monotonic, so we will force that here
    if pixel_dist is not None:
        dist_range = np.arange(pixel_dist.max() + 1)
        smallest_seen = np.inf
        dist_mean_list = []
        for d in dist_range:
            currently_seen = np.mean(pixel_mean[pixel_dist == d])
            if np.isfinite(currently_seen) and currently_seen < smallest_seen:
                smallest_seen = currently_seen
                dist_mean_list.append(currently_seen)
            else:
                dist_mean_list.append(smallest_seen)
        dist_mean = np.array(dist_mean_list)
    else:
        dist_mean = None
        dist_range = None

    # mean_per_bin vs dist_per_bin
    if dist_mean is not None and mean_per_bin is None \
            and dist_per_bin is not None:
        mean_per_bin = np.interp(dist_per_bin, dist_range, dist_mean)
    elif dist_per_bin is None and mean_per_bin is not None:
        dist_per_bin = np.interp(mean_per_bin, dist_mean[::-1],
                                 dist_range[::-1])

    # var_per_bin vs disp_per_bin
    if var_per_bin is None and disp_per_bin is not None:
        var_per_bin = mvr(mean_per_bin, disp_per_bin)
    elif disp_per_bin is None and var_per_bin is not None:
        disp_per_bin = inverse_mvr(mean_per_bin, var_per_bin)

    # pixel_var vs pixel_disp
    if pixel_var is None and pixel_disp is not None:
        pixel_var = mvr(pixel_mean, pixel_disp)
    elif pixel_disp is None and pixel_var is not None:
        pixel_disp = inverse_mvr(pixel_mean, pixel_var)
    else:
        raise ValueError('exactly one of pixel_var and pixel_disp must be '
                         'passed')

    # pixel_var_fit vs pixel_disp_fit
    if pixel_var_fit is None and pixel_disp_fit is not None:
        pixel_var_fit = mvr(pixel_mean, pixel_disp_fit)
    elif pixel_disp_fit is None and pixel_var_fit is not None:
        pixel_disp_fit = inverse_mvr(pixel_mean, pixel_var_fit)

    # convert pixel_* to dist_* quantities part 2: the fitted values
    dist_var_fit = None
    dist_disp_fit = None
    if dist_mean is not None and pixel_disp_fit is not None:
        dist_disp_fit = np.array([np.mean(pixel_disp_fit[pixel_dist == d])
                                  for d in dist_range])
        dist_var_fit = mvr(dist_mean, dist_disp_fit)

    # establish which y values will go into the cloud, bins, and fit
    if yaxis == 'var':
        cloud_y = pixel_var
        bin_y = var_per_bin
        fit_y = dist_var_fit if fit_align_dist else pixel_var_fit
        fit_label = r'$\hat{\sigma}^2$'
        ylabel = 'variance'
    elif yaxis == 'disp':
        cloud_y = pixel_disp
        bin_y = disp_per_bin
        fit_y = dist_disp_fit if fit_align_dist else pixel_disp_fit
        fit_label = r'$\hat{\alpha}$'
        ylabel = 'dispersion'
    else:
        raise ValueError('yaxis must be \'disp\' or \'var\'')

    # establish which x values will go into the cloud, bins, and fit
    if xaxis == 'dist':
        cloud_x = pixel_dist
        bin_x = dist_per_bin
        fit_x = dist_range if fit_align_dist else pixel_dist
        xlabel = 'distance'
    elif xaxis == 'mean':
        cloud_x = pixel_mean
        bin_x = mean_per_bin
        fit_x = dist_mean if fit_align_dist else pixel_mean
        xlabel = 'mean'
    else:
        raise ValueError('yaxis must be \'disp\' or \'var\'')

    # determine reasonable plot limits
    if xaxis == 'mean':
        xmin = mean_min
        xmax = np.percentile(pixel_mean, 99.99)
    elif xaxis == 'dist':
        xmin = 0
        xmax = dist_max
    else:
        raise ValueError('xaxis must be \'mean\' or \'dist\'')

    # prepare and apply filter indexes
    cloud_idx = np.isfinite(cloud_x) & np.isfinite(cloud_y) & \
        ((cloud_y > 0) if logy else True) & \
        ((cloud_x > 0) if logx else True) & \
        (cloud_x >= xmin) & (cloud_x <= xmax)
    cloud_x = cloud_x[cloud_idx]
    cloud_y = cloud_y[cloud_idx]
    try:
        bin_idx = np.isfinite(bin_x) & np.isfinite(bin_y) & \
            ((bin_y > 0) if logy else True) & \
            ((bin_x > 0) if logx else True) & \
            (bin_x >= xmin) & (bin_x <= xmax)
        bin_x = bin_x[bin_idx]
        bin_y = bin_y[bin_idx]
    except TypeError:
        pass
    fit_idx = np.isfinite(fit_x) & np.isfinite(fit_y) & \
        ((fit_y > 0) if logy else True) & \
        ((fit_x > 0) if logx else True) & \
        (fit_x >= xmin) & (fit_x <= xmax)
    fit_x = fit_x[fit_idx]
    fit_y = fit_y[fit_idx]

    if hexbin:
        ymin = max(np.percentile(cloud_y, 0.1), 1e-7) if logy \
            else min(np.percentile(cloud_y, 0.1), 0)
        ymax = np.percentile(cloud_y, 99.9)
    else:
        ymin = None
        ymax = None

    # override the default plot limits above if xlim/ylim were passed
    if xlim is not None:
        xmin, xmax = xlim
    if ylim is not None:
        ymin, ymax = ylim

    # hexbin per-pixel values
    if hexbin:
        scatter(cloud_x, cloud_y, logx=logx, logy=logy, hexbin=True,
                xlim=[xmin, xmax], ylim=[ymin, ymax])

    # scatter per-bin estimates
    if bin_x is not None and bin_y is not None:
        plt.scatter(bin_x, bin_y, label='%s per bin' % fit_label, color='C1',
                    s=scatter_size)

    # plot or scatter fitted estimates
    if fit_x is not None and fit_y is not None:
        sort_idx = np.argsort(fit_x) if not fit_align_dist \
            else np.arange(fit_x.shape[0])
        if scatter_fit > 0:
            plt.scatter(fit_x[sort_idx][::fit_x.shape[0]/scatter_fit],
                        fit_y[sort_idx][::fit_x.shape[0]/scatter_fit],
                        label=fit_label, color='C4', s=scatter_size)
        elif scatter_fit == -1:
            plt.plot(fit_x[sort_idx], fit_y[sort_idx],
                     label='smoothed %s' % fit_label, linewidth=3, color='C4')

    # add poisson line
    if xaxis == 'mean':
        if yaxis == 'var':
            xs = 10**np.linspace(np.log10(xmin), np.log10(xmax), 100) if logx \
                else np.linspace(xmin, xmax, 100)
            plt.plot(xs, xs, label='Poisson', linestyle='--', linewidth=3,
                     color='C3')
        elif yaxis == 'disp':
            plt.hlines(0, xmin, xmax, label='Poisson', linestyle='--',
                       linewidth=3, color='C3')
        else:
            raise ValueError('yaxis must be \'disp\' or \'var\'')
    else:
        if yaxis == 'disp' and not logy:
            plt.hlines(0, xmin, xmax, label='Poisson', linestyle='--',
                       linewidth=3, color='C3')

    # cleanup
    if hexbin:
        plt.ylim((ymin, ymax))
        plt.xlim((xmin, xmax))
    if not hexbin:
        if logx:
            plt.xscale('log')
        if logy:
            plt.yscale('log')
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    if legend:
        plt.legend()


@plotter
def plot_ddr(dist_per_bin, disp_per_bin, disp_fn, scatter_size=36, legend=True,
             **kwargs):
    """
    Simplified plotter to visualized distance-dispersion relationships.

    Plots per-distance dispersion estimates and the fitted dispersion curve.

    Parameters
    ----------
    dist_per_bin, disp_per_bin : np.ndarray
        The distances and estimated dispersions for each distance, respectively.
    disp_fn : function
        The smoothed dispersion function. Returns the smoothed dispersion as a
        function of distance.
    scatter_size : int
        The marker size when plotting scatter plots.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis
        The axis plotted on.
    """
    xmin = dist_per_bin.min()
    xmax = dist_per_bin.max()
    xs = np.arange(xmin, xmax + 1)
    ys = disp_fn(xs)
    plt.scatter(dist_per_bin, disp_per_bin, label=r'$\hat{\alpha}$ per bin',
                color='C1', s=scatter_size)
    plt.plot(xs, ys, label=r'smoothed $\hat{\alpha}$', color='C4', lw=3)
    plt.hlines(0, xmin, xmax, label='Poisson', color='C3', lw=3, linestyle='--')
    plt.ylabel('dispersion')
    plt.xlabel('distance')
    if legend:
        plt.legend()



"""
Experimental module exposing variants of the HiC3DeFDR model for benchmarking
purposes.
"""

import numpy as np
import scipy.stats as stats

from hic3defdr.analysis import HiC3DeFDR
from hic3defdr.util.logging import eprint
from hic3defdr.util.dispersion import mme_per_pixel
from hic3defdr.util.clusters import load_clusters
from hic3defdr.util.parallelization import parallel_apply
import hic3defdr.util.dispersion as dispersion


def poisson_fit_mu_hat(raw, f):
    return np.average(raw / f, weights=f, axis=1)


def poisson_logpmf(x, mu):
    return stats.poisson(mu).logpmf(x)


def poisson_lrt(raw, f, design, refit_mu=True):
    if refit_mu:
        mu_hat_null = poisson_fit_mu_hat(raw, f)
        mu_hat_alt = np.array(
            [poisson_fit_mu_hat(raw[:, design[:, c]], f[:, design[:, c]])
             for c in range(design.shape[1])]
        ).T
    else:
        mu_hat_null = np.mean(raw / f, axis=1)
        mu_hat_alt = np.array(
            [np.mean(raw[:, design[:, c]] / f[:, design[:, c]], axis=1)
             for c in range(design.shape[1])])
    mu_hat_alt_wide = np.dot(mu_hat_alt, design.T)
    null_ll = np.sum(poisson_logpmf(raw, mu_hat_null[:, None] * f), axis=1)
    alt_ll = np.sum(poisson_logpmf(raw, mu_hat_alt_wide * f), axis=1)
    llr = null_ll - alt_ll
    pvalues = stats.chi2(design.shape[1] - 1).sf(-2 * llr)
    return pvalues, llr, mu_hat_null, mu_hat_alt


class Poisson3DeFDR(HiC3DeFDR):
    def estimate_disp(self, estimator='qcml', frac=None, auto_frac_factor=15.,
                      weighted_lowess=True, n_threads=-1):
        # note: all kwargs are ignored
        eprint('estimating dispersion')
        estimator = dispersion.__dict__[estimator] \
            if estimator in dispersion.__dict__ else estimator

        eprint('  loading data')
        disp_idx, _ = self.load_data('disp_idx', 'all')
        row, offsets = self.load_data('row', 'all', idx=disp_idx)
        col, _ = self.load_data('col', 'all', idx=disp_idx)
        scaled, _ = self.load_data('scaled', 'all', idx=disp_idx)

        eprint('  computing pixel-wise mean per condition')
        disp_per_dist = np.zeros((self.dist_thresh_max+1, self.design.shape[1]))
        disp = np.zeros((disp_idx.sum(), self.design.shape[1]))

        def disp_fn(mean):
            return np.zeros_like(mean)

        for c, cond in enumerate(self.design.columns):
            self.save_disp_fn(cond, disp_fn)

        eprint('  saving estimated dispersions to disk')
        self.save_data(disp, 'disp', offsets)
        self.save_data(disp_per_dist, 'disp_per_dist')

    def lrt(self, chrom=None, refit_mu=True, n_threads=-1, verbose=True):
        if chrom is None:
            if n_threads:
                parallel_apply(
                    self.lrt,
                    [{'chrom': c, 'refit_mu': refit_mu, 'verbose': False}
                     for c in self.chroms],
                    n_threads=n_threads
                )
            else:
                for chrom in self.chroms:
                    self.lrt(chrom=chrom, refit_mu=refit_mu)
            return
        eprint('running LRT for chrom %s' % chrom)
        eprint('  loading data', skip=not verbose)
        bias = self.load_bias(chrom)
        size_factors = self.load_data('size_factors', chrom)
        row = self.load_data('row', chrom)
        col = self.load_data('col', chrom)
        raw = self.load_data('raw', chrom)
        disp_idx = self.load_data('disp_idx', chrom)

        eprint('  computing LRT results', skip=not verbose)
        f = bias[row, :][disp_idx, :] * bias[col, :][disp_idx, :] * \
            size_factors[disp_idx, :]
        pvalues, llr, mu_hat_null, mu_hat_alt = poisson_lrt(
            raw[disp_idx, :], f, self.design.values, refit_mu=True)

        if self.loop_patterns:
            eprint('  making loop_idx', skip=not verbose)
            loop_pixels = set().union(
                *sum((load_clusters(pattern.replace('<chrom>', chrom))
                      for pattern in self.loop_patterns.values()), []))
            loop_idx = np.array([True if pixel in loop_pixels else False
                                 for pixel in zip(row[disp_idx],
                                                  col[disp_idx])])
            self.save_data(loop_idx, 'loop_idx', chrom)

        eprint('  saving results to disk', skip=not verbose)
        self.save_data(pvalues, 'pvalues', chrom)
        self.save_data(llr, 'llr', chrom)
        self.save_data(mu_hat_null, 'mu_hat_null', chrom)
        self.save_data(mu_hat_alt, 'mu_hat_alt', chrom)


class Unsmoothed3DeFDR(HiC3DeFDR):
    def estimate_disp(self, estimator='qcml', frac=None, auto_frac_factor=15.,
                      weighted_lowess=True, n_threads=-1):
        # note: all kwargs are ignored
        eprint('estimating dispersion')
        eprint('  loading data')
        disp_idx, _ = self.load_data('disp_idx', 'all')
        row, offsets = self.load_data('row', 'all', idx=disp_idx)
        col, _ = self.load_data('col', 'all', idx=disp_idx)
        scaled, _ = self.load_data('scaled', 'all', idx=disp_idx)

        eprint('  computing pixel-wise mean per condition')
        disp = np.zeros((disp_idx.sum(), self.design.shape[1]))
        for c, cond in enumerate(self.design.columns):
            eprint('  estimating dispersion for condition %s' % cond)
            disp[:, c] = np.maximum(mme_per_pixel(
                scaled[:, self.design[cond]]), 1e-7)

        eprint('  saving estimated dispersions to disk')
        self.save_data(disp, 'disp', offsets)


class Global3DeFDR(HiC3DeFDR):
    def estimate_disp(self, estimator='qcml', frac=None, auto_frac_factor=15.,
                      weighted_lowess=True, n_threads=-1):
        # note: all kwargs except estimator are ignored
        eprint('estimating dispersion')
        estimator = dispersion.__dict__[estimator] \
            if estimator in dispersion.__dict__ else estimator

        eprint('  loading data')
        disp_idx, disp_idx_offsets = self.load_data('disp_idx', 'all')
        loop_idx, _ = self.load_data('loop_idx', 'all')
        row, offsets = self.load_data('row', 'all', idx=disp_idx)
        col, _ = self.load_data('col', 'all', idx=disp_idx)
        raw, _ = self.load_data('raw', 'all', idx=disp_idx)
        f = np.ones_like(raw, dtype=float)
        for i, chrom in enumerate(self.chroms):
            chrom_slice = slice(offsets[i], offsets[i+1])
            row_chrom = row[chrom_slice]
            col_chrom = col[chrom_slice]
            disp_idx_chrom = disp_idx[disp_idx_offsets[i]:disp_idx_offsets[i+1]]
            bias = self.load_bias(chrom)
            size_factors = self.load_data('size_factors', chrom)[disp_idx_chrom]
            f[chrom_slice] = bias[row_chrom, :] * bias[col_chrom, :] \
                * size_factors

        disp = np.zeros((disp_idx.sum(), self.design.shape[1]))
        disp_per_dist = np.zeros((self.dist_thresh_max+1, self.design.shape[1]))
        for c, cond in enumerate(self.design.columns):
            eprint('  estimating dispersion for condition %s' % cond)
            global_disp = estimator(raw[loop_idx, :][:, self.design[cond]],
                                    f=f[loop_idx, :][:, self.design[cond]])
            disp[:, c] = global_disp
            disp_per_dist[:, c] = global_disp

            def disp_fn(mean):
                return np.ones_like(mean) * global_disp

            self.save_disp_fn(cond, disp_fn)

        eprint('  saving estimated dispersions to disk')
        self.save_data(disp, 'disp', offsets)
        self.save_data(disp_per_dist, 'disp_per_dist')



import sys

import numpy as np
import scipy.sparse as sparse

from hiclite.steps.filter import filter_sparse_rows_count
from hiclite.steps.balance import kr_balance

from hic3defdr import HiC3DeFDR
from hic3defdr import plot_roc, plot_fdr

from hic3defdr.alternatives import Poisson3DeFDR, Unsmoothed3DeFDR, Global3DeFDR


repnames = ['ES_1', 'ES_3', 'NPC_2', 'NPC_4']
#chroms = ['chr%i' % i for i in range(1, 20)] + ['chrX']
chroms = ['chr18', 'chr19']
base_path = 'c:/lab-data/bonev/'
h = HiC3DeFDR(
    [base_path + '<rep>/<chrom>_raw.npz'.replace('<rep>', repname)
     for repname in repnames],
    [base_path + '<rep>/<chrom>_kr.bias'.replace('<rep>', repname)
     for repname in repnames],
    chroms, 'design.csv', 'output-small',
    loop_patterns={c: base_path + 'clusters_0.025/%s_<chrom>_clusters.json' % c
                   for c in ['ES', 'NPC']},
    dist_thresh_max=500,
    mean_thresh=5.0
)
h = HiC3DeFDR.load('output-small')
#h.run_to_qvalues()
#h.prepare_data()
#h.estimate_disp()
#h.lrt()
#h.bh()
#h.classify()
#h.plot_dispersion_fit('ES', xaxis='dist', yaxis='disp', logx=False, logy=False, outfile='test_ddr.png')
#h.plot_dispersion_fit('ES', xaxis='dist', yaxis='var', logx=False, outfile='test_dvr.png')
#h.plot_dispersion_fit('ES', xaxis='mean', yaxis='disp', logy=False, outfile='test_mdr.png')
#h.plot_dispersion_fit('ES', xaxis='mean', yaxis='var', outfile='test_mvr.png')
#for d in [5, 10, 25, 50, 75, 100]:
#    h.plot_dispersion_fit('ES', xaxis='mean', yaxis='var', distance=d, logx=True,
#                          xlim=(5, 500), ylim=(1e-5, 1e3),
#                          outfile='mvr_es_%i.png' % d)
#h.plot_dispersion_fit('ES', ylim=(0, 0.25), outfile='ddr_real.png')


for r in ['all', 'short', 'mid', 'long']:
    loop_pattern = 'ES_<chrom>_%s_clusters.json' % r
    #h.simulate('ES', loop_pattern=loop_pattern, outdir='sim-small-%s' % r)

    infile_pattern = 'sim-small-%s/<rep>_<chrom>_raw.npz' % r
    repnames = ['A1', 'A2', 'B1', 'B2']
    if False:
        for repname in repnames:
            for chrom in chroms:
                sys.stderr.write('balancing rep %s chrom %s\n' % (repname, chrom))
                infile = infile_pattern.replace('<rep>', repname)\
                    .replace('<chrom>', chrom)
                outfile = infile.replace('_raw.npz', '_kr.bias')
                _, bias, _ = kr_balance(
                    filter_sparse_rows_count(sparse.load_npz(infile)), fl=0)
                np.savetxt(outfile, bias)

    base_path = 'sim-small-%s/' % r
    h_sim = HiC3DeFDR(
        [base_path + '<rep>_<chrom>_raw.npz'.replace('<rep>', repname)
         for repname in repnames],
        [base_path + '<rep>_<chrom>_kr.bias'.replace('<rep>', repname)
         for repname in repnames],
        chroms, base_path + 'design.csv', 'output-sim-small-%s' % r,
        loop_patterns={'ES': loop_pattern}
    )
    h_sim = HiC3DeFDR.load('output-sim-small-%s' % r)
    #h_sim.run_to_qvalues()
    #h_sim.plot_dispersion_fit('A', ylim=(0, 0.25), outfile='ddr_sim.png')
    #h_sim.plot_pvalue_distribution(outfile='null_sim_pvalues.png')
    #h_sim.evaluate('ES', 'sim-small-%s/labels_<chrom>.txt' % r)

    h_sim_pois = Poisson3DeFDR(
        [base_path + '<rep>_<chrom>_raw.npz'.replace('<rep>', repname)
         for repname in repnames],
        [base_path + '<rep>_<chrom>_kr.bias'.replace('<rep>', repname)
         for repname in repnames],
        chroms, base_path + 'design.csv', 'output-sim-small-%s-pois' % r,
        loop_patterns={'ES': loop_pattern}
    )
    h_sim_pois = Poisson3DeFDR.load('output-sim-small-%s-pois' % r)
    #h_sim_pois.run_to_qvalues()
    #h_sim_pois.evaluate('ES', 'sim-small-%s/labels_<chrom>.txt' % r)

    h_sim_unsmooth = Unsmoothed3DeFDR(
        [base_path + '<rep>_<chrom>_raw.npz'.replace('<rep>', repname)
         for repname in repnames],
        [base_path + '<rep>_<chrom>_kr.bias'.replace('<rep>', repname)
         for repname in repnames],
        chroms, base_path + 'design.csv', 'output-sim-small-%s-unsmooth' % r,
        loop_patterns={'ES': loop_pattern}
    )
    h_sim_unsmooth = Unsmoothed3DeFDR.load('output-sim-small-%s-unsmooth' % r)
    #h_sim_unsmooth.run_to_qvalues()
    #h_sim_unsmooth.evaluate('ES', 'sim-small-%s/labels_<chrom>.txt' % r)

    h_sim_global = Global3DeFDR(
        [base_path + '<rep>_<chrom>_raw.npz'.replace('<rep>', repname)
         for repname in repnames],
        [base_path + '<rep>_<chrom>_kr.bias'.replace('<rep>', repname)
         for repname in repnames],
        chroms, base_path + 'design.csv', 'output-sim-small-%s-global' % r,
        loop_patterns={'ES': loop_pattern}
    )
    h_sim_global = Global3DeFDR.load('output-sim-small-%s-global' % r)
    #h_sim_global.run_to_qvalues()
    #h_sim_global.evaluate('ES', 'sim-small-%s/labels_<chrom>.txt' % r)

    plot_roc([np.load('output-sim-small-%s/eval.npz' % r),
              np.load('output-sim-small-%s-pois/eval.npz' % r),
              np.load('output-sim-small-%s-unsmooth/eval.npz' % r),
              np.load('output-sim-small-%s-global/eval.npz' % r)],
             ['3DeFDR-HiC', 'Poisson', 'Sample Variance', 'Global NB'],
             outfile='roc_%s.png' % r)

    plot_fdr([np.load('output-sim-small-%s/eval.npz' % r),
              np.load('output-sim-small-%s-pois/eval.npz' % r),
              np.load('output-sim-small-%s-unsmooth/eval.npz' % r),
              np.load('output-sim-small-%s-global/eval.npz' % r)],
             ['3DeFDR-HiC', 'Poisson', 'Sample Variance', 'Global NB'],
             outfile='fdr_%s.png' % r)

    plot_roc([np.load('output-sim-small-%s/eval.npz' % r),
              np.load('output-sim-small-%s-unsmooth/eval.npz' % r)],
             ['3DeFDR-HiC', 'Sample Variance'],
             outfile='roc_twolines_%s.png' % r)

    plot_fdr([np.load('output-sim-small-%s/eval.npz' % r),
              np.load('output-sim-small-%s-unsmooth/eval.npz' % r)],
             ['3DeFDR-HiC', 'Sample Variance'],
             outfile='fdr_twolines_%s.png' % r)
#f.process_chrom('chr19')
#f.process_all()
#f.bh()
#f.threshold_chrom('chr19')
#f.threshold_all()
#base_path = 'c:/lab-data/bonev/'
#f.simulate('ES', base_path + 'clusters/ES_<chrom>_clusters.json', chrom='chr19')
#f.simulate('ES', base_path + 'clusters/ES_<chrom>_clusters.json')

#from hic3defdr import HiC3DeFDR
#f = HiC3DeFDR.load('output')
#f.estimate_disp('chr19')
#f.estimate_disp('chrX')
#f.prepare_data('chrX')
#f.plot_dispersion_fit('chr19', 'ES', outfile='disp.png')
#f.plot_dispersion_fit('chr19', 'ES', yaxis='var', outfile='var.png')
#f.plot_qvalue_distribution(outfile='qvalue_dist.png')
#f.process_chrom('chr19')
#f.lrt('chr19')

#f.threshold_chrom('chr19')

#f.plot_grid('chr1', 1303, 1312, 50, outfile='grid.png')
#f.plot_dispersion_fit('chr19', 'ES', outfile='disp_es_2.png')
#f.plot_dispersion_fit('chr19', 'NPC', outfile='disp_npc_2.png')
#f.classify_sig_pixels('chr19')


#import matplotlib.pyplot as plt
#_, ax, outline_clusters = f.plot_grid('chr1', 1303, 1312, 50)
#ax[0, 2].plot([0, 100], [0, 100])
#plt.savefig('grid.png', bbox_inches='tight')
#plt.clf()

#cluster_idx = np.array([True if any((r, c in cluster for cluster in clusters)) else False
#                        for r, c in zip(row[disp_idx][loop_idx], col[disp_idx][loop_idx])])

#from hic3defdr import HiC3DeFDR
#f = HiC3DeFDR.load('output')
#f.plot_dispersion_fit('chr19', 'ES', outfile='disp_es.png')

#import numpy as np
#import scipy.stats as stats
#from hic3defdr import HiC3DeFDR
#from hic3defdr.scaled_nb import inverse_mvr
#f = HiC3DeFDR.load('output')
#cond = 'ES'
#chrom = 'chr1'
#disp_idx = f.load_data('disp_idx', 'chr1')
#row = f.load_data('row', chrom)[disp_idx]
#col = f.load_data('col', chrom)[disp_idx]
#scaled = f.load_data('scaled', chrom)[disp_idx, :][:, f.design[cond]]
#mean = np.mean(scaled, axis=1)
#var = np.var(scaled, ddof=1, axis=1)
#disp = inverse_mvr(mean, var)
#dist = col - row
#stats.spearmanr(mean, disp)
#stats.spearmanr(dist, disp)



from __future__ import print_function
import sys


def eprint(*args, **kwargs):
    """
    Drop-in replacement for ``print()`` that prints to stderr instead of stdout.

    Parameters
    ----------
    args, kwargs : args, kwargs
        All args and kwargs will be passed through to ``print()`` except for the
        special kwarg ``skip``; if this kwarg is present and set to True,
        nothing will be printed.
    """
    if not kwargs.get('skip', False):
        if 'skip' in kwargs:
            del kwargs['skip']
        print(*args, file=sys.stderr, **kwargs)



import numpy as np
import scipy.sparse as sparse

from hic3defdr.util.clusters import find_clusters


def threshold_and_cluster(qvalues, row, col, fdr):
    """
    Thresholds pixels by comparing their q-values to a target FDR and clusters
    the significant and insignificant pixels.

    Parameters
    ----------
    qvalues : np.ndarray
        The qvalue for each pixel under consideration.
    row, col : np.ndarray
        The row and column indices corresponding to the qvalues.
    fdr : float
        The FDR to threshold on.

    Returns
    -------
    sig_clusters, insig_clusters : list of set of tuple of int
        Lists of the significant and insignificant clusters, respectively.
    """
    # threshold on FDR
    sig_idx = qvalues < fdr
    insig_idx = qvalues >= fdr

    # gather and cluster sig and insig points
    n = max(row.max(), col.max()) + 1  # guess matrix shape
    sig_points = sparse.coo_matrix(
        (np.ones(sig_idx.sum(), dtype=bool),
         (row[sig_idx],
          col[sig_idx])), shape=(n, n))
    insig_points = sparse.coo_matrix(
        (np.ones(insig_idx.sum(), dtype=bool),
         (row[insig_idx],
          col[insig_idx])), shape=(n, n))
    sig_clusters = find_clusters(sig_points)
    insig_clusters = find_clusters(insig_points)
    return sig_clusters, insig_clusters


def size_filter(clusters, cluster_size):
    """
    Filters out clusters which are smaller than cluster_size.

    Parameters
    ----------
    clusters : list of set of tuple of int
        The clusters to filter.
    cluster_size : int
        The minimum size of a cluster needed to pass this filter.

    Returns
    -------
    list of set of tuple of int
        The filtered clusters.
    """
    return [c for c in clusters if len(c) >= cluster_size]



import numpy as np
import scipy.stats as stats

from hic3defdr.util.scaled_nb import logpmf, fit_mu_hat


def lrt(raw, f, disp, design, refit_mu=True):
    """
    Performs a likelihood ratio test on raw data ``raw`` given scaling factors
    ``f`` and dispersion ``disp``.

    Parameters
    ----------
    raw, f, disp : np.ndarray
        Matrices of raw values, combined scaling factors, and dispersions,
        respectively. Rows correspond to pixels, columns correspond to
        replicates.
    design : np.ndarray
        Describes the grouping of replicates into conditions. Rows correspond to
        replicates, columns correspond to conditions, and values should be True
        where a replicate belongs to a condition and False otherwise.

    Returns
    -------
    pvalues : np.ndarray
        The LRT p-values per pixel.
    llr : np.ndarray
        The log likelihood ratio per pixel.
    mu_hat_null, mu_hat_alt : np.ndarray
        The fitted mean parameters under the null and alt models, respectively,
        per pixel.
    """
    if refit_mu:
        mu_hat_null = fit_mu_hat(raw, f, disp)
        mu_hat_alt = np.array(
            [fit_mu_hat(raw[:, design[:, c]],
                        f[:, design[:, c]],
                        disp[:, design[:, c]])
             for c in range(design.shape[1])]).T
    else:
        mu_hat_null = np.mean(raw / f, axis=1)
        mu_hat_alt = np.array(
            [np.mean(raw[:, design[:, c]] / f[:, design[:, c]], axis=1)
             for c in range(design.shape[1])]).T
    mu_hat_alt_wide = np.dot(mu_hat_alt, design.T)
    null_ll = np.sum(logpmf(raw, mu_hat_null[:, None] * f, disp), axis=1)
    alt_ll = np.sum(logpmf(raw, mu_hat_alt_wide * f, disp), axis=1)
    llr = null_ll - alt_ll
    pvalues = stats.chi2(design.shape[1] - 1).sf(-2 * llr)
    return pvalues, llr, mu_hat_null, mu_hat_alt



import numpy as np
import matplotlib.pyplot as plt

from lib5c.util.plotting import plotter


@plotter
def plot_dd_curves(row, col, before, after, repnames=None, log=True, **kwargs):
    """
    Plots a comparison of distance dependence curves before and after scaling.

    Parameters
    ----------
    row, col : np.ndarray
        Row and column indices identifying the location of pixels in ``before``
        and ``after``.
    before, after: np.ndarray
        Counts per pixel before and after scaling, respectively.
    repnames : list of str, optional
        Pass the replicate names in the same order as the columns of ``before``
        and ``after``.
    log : bool
        Pass True to log both axes of the distance dependence plot.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis
        The axis plotted on.
    """
    if repnames is None:
        repnames = ['%s' % (i+1) for i in range(before.shape[1])]
    dist = col - row
    n = max(row.max(), col.max()) + 1
    dist_bin_idx = np.digitize(dist, np.linspace(0, 1000, 101), right=True)
    bs = np.arange(1, dist_bin_idx.max() + 1)
    before_means = np.array(
        [np.sum(before[dist_bin_idx == b, :], axis=0) / float(n - b)
         for b in bs])
    after_means = np.array(
        [np.sum(after[dist_bin_idx == b, :], axis=0) / float(n - b)
         for b in bs])
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 5))
    for r, repname in enumerate(repnames):
        ax1.plot(bs*10 + 5, before_means[:, r], label=repname, color='C%i' % r)
        ax2.plot(bs*10 + 5, after_means[:, r], label=repname, color='C%i' % r)
    plt.legend()
    ax1.set_xlabel('distance (bins)')
    ax2.set_xlabel('distance (bins)')
    ax1.set_ylabel('average counts')
    ax1.set_title('before scaling')
    ax2.set_title('after scaling')
    if log:
        ax1.set_yscale('log')
        ax1.set_xscale('log')
        ax2.set_yscale('log')
        ax2.set_xscale('log')



import sys

import numpy as np
import scipy.sparse as sparse

from hiclite.steps.filter import filter_sparse_rows_count
from hiclite.steps.balance import kr_balance

from hic3defdr import HiC3DeFDR


repnames = ['ES_1', 'ES_3', 'NPC_2', 'NPC_4']
chroms = ['chr19']
base_path = 'c:/lab-data/bonev/'

f = HiC3DeFDR(
    [base_path + '<rep>/<chrom>_raw.npz'.replace('<rep>', repname)
     for repname in repnames],
    [base_path + '<rep>/<chrom>_kr.bias'.replace('<rep>', repname)
     for repname in repnames],
    chroms, 'design.csv', 'output-test',
    loop_patterns={c: base_path + 'clusters_0.025/%s_<chrom>_clusters.json' % c
                   for c in ['ES', 'NPC']}
)
f.run_to_qvalues()
f.simulate('ES', beta=0.0)

infile_pattern = 'sim/<rep>_<chrom>_raw.npz'
repnames = ['A1', 'A2', 'B1', 'B2']
sim_path = 'sim/'

for repname in repnames:
    for chrom in chroms:
        sys.stderr.write('balancing rep %s chrom %s\n' % (repname, chrom))
        infile = infile_pattern.replace('<rep>', repname)\
            .replace('<chrom>', chrom)
        outfile = infile.replace('_raw.npz', '_kr.bias')
        _, bias, _ = kr_balance(
            filter_sparse_rows_count(sparse.load_npz(infile)), fl=0)
        np.savetxt(outfile, bias)

f_sim_cml = HiC3DeFDR(
    raw_npz_patterns=[sim_path + '<rep>_<chrom>_raw.npz'.replace('<rep>', repname) for repname in repnames],
    bias_patterns=[sim_path + '<rep>_<chrom>_kr.bias'.replace('<rep>', repname) for repname in repnames],
    chroms=chroms,
    design=sim_path + 'design.csv',
    outdir='output-cml',
    loop_patterns={'ES': base_path + 'clusters_0.025/ES_<chrom>_clusters.json'}
)
f_sim_cml.run_to_qvalues(estimator='cml')
f_sim_cml.plot_pvalue_distribution(outfile='pvalues_cml.png')

#f_sim_qcml = HiC3DeFDR(
#    raw_npz_patterns=[sim_path + '<rep>_<chrom>_raw.npz'.replace('<rep>', repname) for repname in repnames],
#    bias_patterns=[sim_path + '<rep>_<chrom>_kr.bias'.replace('<rep>', repname) for repname in repnames],
#    chroms=chroms,
#    design=sim_path + 'design.csv',
#    outdir='output-qcml',
#    loop_patterns={'ES': base_path + 'clusters_0.025/ES_<chrom>_clusters.json'}
#)
#f_sim_qcml.run_to_qvalues(estimator='qcml')
#f_sim_qcml.plot_pvalue_distribution(outfile='pvalues_qcml.png')
#
#f_sim_mme = HiC3DeFDR(
#    raw_npz_patterns=[sim_path + '<rep>_<chrom>_raw.npz'.replace('<rep>', repname) for repname in repnames],
#    bias_patterns=[sim_path + '<rep>_<chrom>_kr.bias'.replace('<rep>', repname) for repname in repnames],
#    chroms=chroms,
#    design=sim_path + 'design.csv',
#    outdir='output-mme',
#    loop_patterns={'ES': base_path + 'clusters_0.025/ES_<chrom>_clusters.json'}
#)
#f_sim_mme.run_to_qvalues(estimator='mme')
#f_sim_mme.plot_pvalue_distribution(outfile='pvalues_mme.png')
#
#f_sim_mvr = HiC3DeFDR(
#    raw_npz_patterns=[sim_path + '<rep>_<chrom>_raw.npz'.replace('<rep>', repname) for repname in repnames],
#    bias_patterns=[sim_path + '<rep>_<chrom>_kr.bias'.replace('<rep>', repname) for repname in repnames],
#    chroms=chroms,
#    design=sim_path + 'design.csv',
#    outdir='output-mvr',
#    loop_patterns={'ES': base_path + 'clusters_0.025/ES_<chrom>_clusters.json'}
#)
#f_sim_mvr.run_to_qvalues(estimator='mvr')
#f_sim_mvr.plot_pvalue_distribution(outfile='pvalues_mvr.png')



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from lib5c.util.plotting import plotter


@plotter
def plot_fn_vs_fp(eval_results, labels, threshold=0.15, colors=None,
                  xlabel='label', **kwargs):
    """
    Plots two bar plots, one showing FNR and the other showing FPR at a fixed
    threshold.

    Parameters
    ----------
    eval_results : list of dict-like or list of list of dict-like
        The dicts should have keys 'thresh', 'fpr', and 'tpr' whose values are
        parallel vectors describing the thresholds, FPRs, and TPRs to use for
        the bar plots. Each dict in the list represents a different bar which
        will be added to each bar plot. Pass a nested list of dicts to draw
        grouped bar plots, denoting the outer list by color and the inner list
        by x-axis position.
    labels : list of str or (list of str, list of str)
        List of labels parallel to ``eval_results`` providing names for each
        bar. If ``eval_results`` is a nested list, pass a tuple of two lists.
        The first list should provide the labels for the outer list grouping
        (``len(labels[0]) == len(eval_results)``) and the second should provide
        the labels for the inner list grouping
        (``len(labels[1]) == len(eval_results[i])`` for any ``i``).
    threshold : float
        The fixed threshold at which the FNR and FPR will be plotted. In
        practice, this function will use the closest threshold found in each
        dict in ``eval_results``.
    colors : matplotlib color or list of colors, optional
        Specify the color to use for the bars in the bar plot. If
        ``eval_results`` is a nested list, pass a list of colors to color core
        the outer list grouping (``len(colors) == len(eval_results)``). Pass
        None to use automatic colors.
    xlabel : str
        The label to use for the x-axis.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis, array of pyplot axes
        The first pyplot axis returned is injected by ``@plotter``. The array of
        pyplot axes is the second return value from the call to
        ``plt.subplots()`` that is used to create the pair of barplots.
    """
    data = []
    if type(eval_results[0]) in [list, tuple]:
        hue = 'group'
        color = None
        palette = colors if colors else None
        for res_group, group_label in zip(eval_results, labels[0]):
            for res, label in zip(res_group, labels[1]):
                if res is None:
                    continue
                # unbox results
                fpr = res['fpr']
                fnr = 1 - res['tpr']
                thresh = 1 - res['thresh']

                # find closest thresh
                idx = np.argmin(np.abs(thresh - threshold))

                # append to data
                data.append({xlabel: label, 'group': group_label,
                             'FPR': fpr[idx], 'FNR': fnr[idx]})
    else:
        hue = None
        color = colors if colors else 'k'
        palette = None
        for res, label in zip(eval_results, labels):
            # unbox results
            fpr = res['fpr']
            fnr = 1 - res['tpr']
            thresh = 1 - res['thresh']

            # find closest thresh
            idx = np.argmin(np.abs(thresh - threshold))

            # append to data
            data.append({xlabel: label, 'FPR': fpr[idx], 'FNR': fnr[idx]})
    df = pd.DataFrame(data)
    fig, axes = plt.subplots(1, 2, figsize=(6, 3))
    plt.subplots_adjust(wspace=0.4)
    sns.barplot(data=df, x=xlabel, y='FNR', hue=hue, palette=palette,
                color=color, ax=axes[0])
    sns.barplot(data=df, x=xlabel, y='FPR', hue=hue, palette=palette,
                color=color, ax=axes[1])
    if hue is not None:
        axes[0].legend_.remove()
        axes[1].legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    return axes



from hic3defdr.analysis import HiC3DeFDR
from hic3defdr.plotting.roc import plot_roc
from hic3defdr.plotting.fdr import plot_fdr
from hic3defdr.plotting.fn_vs_fp import plot_fn_vs_fp
from hic3defdr.plotting.distance_bias import plot_distance_bias
from hic3defdr.plotting.dispersion import compare_disp_fits

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

__all__ = [
    'HiC3DeFDR',
    'plot_roc',
    'plot_fdr',
    'plot_fn_vs_fp',
    'plot_distance_bias',
    'compare_disp_fits'
]



from hic3defdr import HiC3DeFDR

repnames = ['ES_1', 'ES_3', 'NPC_2', 'NPC_4']
#chroms = ['chr%i' % i for i in range(1, 20)] + ['chrX']
chroms = ['chr19']
base_path = 'c:/lab-data/bonev/'
f = HiC3DeFDR(
    [base_path + '<rep>/<chrom>_raw.npz'.replace('<rep>', repname)
     for repname in repnames],
    [base_path + '<rep>/<chrom>_kr.bias'.replace('<rep>', repname)
     for repname in repnames],
    chroms, 'design.csv', 'output-test',
    loop_patterns={c: base_path + 'clusters/%s_<chrom>_clusters.json' % c
                   for c in ['ES', 'NPC']}
)
f = HiC3DeFDR.load('output')
f.run_to_qvalues()
f.classify()
f.process_chrom('chr19')
f.process_all()
f.bh()
f.threshold_chrom('chr19')
f.threshold_all()
base_path = 'c:/lab-data/bonev/'
f.simulate('ES', base_path + 'clusters/ES_<chrom>_clusters.json', chrom='chr19')
f.simulate('ES', base_path + 'clusters/ES_<chrom>_clusters.json')

from hic3defdr import HiC3DeFDR
f = HiC3DeFDR.load('output')
f.estimate_disp('chr19')
f.estimate_disp('chrX')
f.prepare_data('chrX')
f.plot_dispersion_fit('chr19', 'ES', outfile='disp.png')
f.plot_dispersion_fit('chr19', 'ES', yaxis='var', outfile='var.png')
#f.plot_qvalue_distribution(outfile='qvalue_dist.png')
#f.process_chrom('chr19')
f.lrt('chr19')

f.threshold_chrom('chr19')

f.plot_grid('chr1', 1303, 1312, 50, outfile='grid.png')
f.plot_dispersion_fit('chr19', 'ES', outfile='disp_es_2.png')
f.plot_dispersion_fit('chr19', 'NPC', outfile='disp_npc_2.png')
f.classify_sig_pixels('chr19')


import matplotlib.pyplot as plt
_, ax, outline_clusters = f.plot_grid('chr1', 1303, 1312, 50)
ax[0, 2].plot([0, 100], [0, 100])
plt.savefig('grid.png', bbox_inches='tight')
plt.clf()

cluster_idx = np.array([True if any((r, c in cluster for cluster in clusters)) else False
                        for r, c in zip(row[disp_idx][loop_idx], col[disp_idx][loop_idx])])

from hic3defdr import HiC3DeFDR
f = HiC3DeFDR.load('output')
f.plot_dispersion_fit('chr19', 'ES', outfile='disp_es.png')

import numpy as np
import scipy.stats as stats
from hic3defdr import HiC3DeFDR
from hic3defdr.scaled_nb import inverse_mvr
f = HiC3DeFDR.load('output')
cond = 'ES'
chrom = 'chr1'
disp_idx = f.load_data('disp_idx', 'chr1')
row = f.load_data('row', chrom)[disp_idx]
col = f.load_data('col', chrom)[disp_idx]
scaled = f.load_data('scaled', chrom)[disp_idx, :][:, f.design[cond]]
mean = np.mean(scaled, axis=1)
var = np.var(scaled, ddof=1, axis=1)
disp = inverse_mvr(mean, var)
dist = col - row
stats.spearmanr(mean, disp)
stats.spearmanr(dist, disp)



from hic3defdr import HiC3DeFDR

repnames = ['A1', 'A2', 'B1', 'B2']
#chroms = ['chr%i' % i for i in range(1, 20)] + ['chrX']
chroms = ['chr19']
sim_path = 'sim/'
base_path = 'c:/lab-data/bonev/'
f = HiC3DeFDR(
    [sim_path + '<rep>_<chrom>_raw.npz'.replace('<rep>', repname)
     for repname in repnames],
    [sim_path + '<rep>_<chrom>_kr.bias'.replace('<rep>', repname)
     for repname in repnames],
    chroms, sim_path + 'design.csv', 'output-sim',
    loop_patterns={'ES': base_path + 'clusters_0.025/ES_<chrom>_clusters.json'}
)

from hic3defdr import HiC3DeFDR
f = HiC3DeFDR.load('output-sim')
f.evaluate('ES', 'sim/labels_<chrom>.txt')


import numpy as np
from hic3defdr import plot_roc, plot_fdr
plot_roc([np.load('output-sim/eval.npz')], ['hic3defdr'], outfile='roc.png')
plot_fdr([np.load('output-sim/eval.npz')], ['hic3defdr'], outfile='fdr.png')



import os

import numpy as np
import pandas as pd
import scipy.sparse as sparse

from lib5c.util.system import check_outdir
from lib5c.util.statistics import adjust_pvalues

from hic3defdr.util.logging import eprint
from hic3defdr.util.clusters import load_clusters
from hic3defdr.util.simulation import simulate
from hic3defdr.util.evaluation import make_y_true, evaluate
from hic3defdr.util.progress import tqdm_maybe as tqdm
from hic3defdr.util.parallelization import parallel_apply


class SimulatingHiC3DeFDR(object):
    """
    Mixin class containing plotting functions for HiC3DeFDR.
    """
    def simulate(self, cond, chrom=None, beta=0.5, p_diff=0.4, skip_bias=False,
                 loop_pattern=None, outdir='sim', n_threads=-1, verbose=True):
        """
        Simulates raw contact matrices based on previously fitted scaled means
        and dispersions in a specific condition.

        Can only be run after ``estimate_dispersions()`` has been run.

        Parameters
        ----------
        cond : str
            Name of the condition to base the simulation on.
        chrom : str, optional
            Name of the chromosome to simulate. Pass None to simulate all
            chromosomes in series.
        beta : float
            The effect size of the loop perturbations to use when simulating.
            Perturbed loops will be strengthened or weakened by this fraction of
            their original strength.
        p_diff : float or list of float
            Pass a single float to specify the probability that a loop will be
            perturbed across the simulated conditions. Pass four floats to
            specify the probabilities of all four specific perturbations: up in
            A, down in A, up in B, down in B. The remaining loops will be
            constitutive.
        skip_bias : bool
            Pass True to set all bias factors and size factors to 1,
            effectively simulating "unbiased" raw data.
        loop_pattern : str, optional
            File path pattern to sparse JSON formatted cluster files
            representing loop cluster locations for the simulation. Should
            contain at least one '<chrom>' which will be replaced with the
            chromosome name when loading data for specific chromosomes. Pass
            None to use ``self.loop_patterns[cond]``.
        outdir : str
            Path to a directory to store the simulated data to.
        n_threads : int
            The number of threads (technically GIL-avoiding child processes) to
            use to process multiple chromosomes in parallel. Pass -1 to use as
            many threads as there are CPUs. Pass 0 to process the chromosomes
            serially.
        verbose : bool
            Pass False to silence reporting of progress to stderr.
        """
        if chrom is None:
            if n_threads:
                parallel_apply(
                    self.simulate,
                    [{'cond': cond, 'chrom': c, 'beta': beta, 'p_diff': p_diff,
                      'skip_bias': skip_bias, 'loop_pattern': loop_pattern,
                      'outdir': outdir, 'verbose': False}
                     for c in self.chroms],
                    n_threads=n_threads
                )
            else:
                for chrom in self.chroms:
                    self.simulate(cond, chrom=chrom, beta=beta, p_diff=p_diff,
                                  loop_pattern=loop_pattern, outdir=outdir)
            return
        eprint('simulating data for chrom %s' % chrom)
        # resolve loop_pattern
        if loop_pattern is None:
            loop_pattern = self.loop_patterns[cond]

        # load everything
        bias = self.load_bias(chrom)[:, self.design[cond]]
        size_factors = self.load_data('size_factors', chrom)
        if len(size_factors.shape) == 2:
            size_factors = size_factors[:, self.design[cond]]
        else:
            size_factors = size_factors[self.design[cond]]
        row = self.load_data('row', chrom)
        col = self.load_data('col', chrom)
        scaled = self.load_data('scaled', chrom)[:, self.design[cond]]
        disp_fn = self.load_disp_fn(cond)
        clusters = load_clusters(loop_pattern.replace('<chrom>', chrom))

        # compute pixel-wise mean of normalized data
        mean = np.mean(scaled, axis=1)

        # book keeping
        check_outdir('%s/' % outdir)
        n_sim = size_factors.shape[-1] * 2
        repnames = sum((['%s%i' % (c, i+1) for i in range(n_sim/2)]
                        for c in ['A', 'B']), [])

        # write design to disk if not present
        design_file = '%s/design.csv' % outdir
        if not os.path.isfile(design_file):
            pd.DataFrame(
                {'A': [1]*(n_sim/2) + [0]*(n_sim/2),
                 'B': [0]*(n_sim/2) + [1]*(n_sim/2)},
                dtype=bool,
                index=repnames
            ).to_csv(design_file)

        # rewrite size_factor matrix in terms of distance
        if len(size_factors.shape) == 2:
            eprint('  converting size factors', skip=not verbose)
            dist = col - row
            n_dists = dist.max() + 1
            new_size_factors = np.zeros((n_dists, size_factors.shape[1]))
            for d in tqdm(range(n_dists)):
                idx = np.argmax(dist == d)
                new_size_factors[d, :] = size_factors[idx, :]
            size_factors = new_size_factors

        # get rid of bias
        if skip_bias:
            bias = np.ones_like(bias)
            size_factors = np.ones_like(size_factors)

        # tile bias and size_factors
        bias = np.tile(bias, 2)
        size_factors = np.tile(size_factors, 2)

        # simulate and save
        classes, sim_iter = simulate(
            row, col, mean, disp_fn, bias, size_factors, clusters, beta=beta,
            p_diff=p_diff, trend='dist', verbose=verbose)
        np.savetxt('%s/labels_%s.txt' % (outdir, chrom), classes, fmt='%s')
        for rep, csr in zip(repnames, sim_iter):
            sparse.save_npz('%s/%s_%s_raw.npz' % (outdir, rep, chrom), csr)

    def evaluate(self, cluster_pattern, label_pattern, min_dist=None,
                 max_dist=None, rerun_bh=False, outfile=None):
        """
        Evaluates the results of this analysis, comparing it to true labels.

        Parameters
        ----------
        cluster_pattern : str
            File path pattern to sparse JSON formatted cluster files
            representing loop cluster locations. Should contain at least one
            '<chrom>' which will be replaced with the chromosome name when
            loading data for specific chromosomes. Pass a condition name to use
            ``self.loop_patterns[cluster_pattern]`` instead.
        label_pattern : str
            File path pattern to true label files for each chromosome. Should
            contain at least one '<chrom>' which will be replaced with the
            chromosome name when loading data for specific chromosomes. Files
            should be loadable with ``np.loadtxt(..., dtype='|S7')`` to yield a
            vector of true labels parallel to the clusters pointed to by
            ``cluster_pattern``.
        min_dist, max_dist : int, optional
            Specify minimum and maximum distances to evaluate performance
            within, respectively. Pass None to leave one or both ends unbounded.
        rerun_bh : bool
            If ``min_dist`` and/or ``max_dist`` are used to constrain the
            distances, pass True to re-run BH-FDR on the subset of p-values at
            the selected distances. Pass False to use the original dataset-wide
            q-values. Does nothing if ``min_dist`` and ``max_dist`` are both
            None.
        outfile : str, optional
            Name of a file to save the evaluation results to inside this
            object's ``outdir``. Default is 'eval.npz' if ``min_dist`` and
            ``max_dist`` are both None, otherwise it is
            'eval_<min_dist>_<max_dist>.npz'.
        """
        # resolve outfile
        if outfile is None:
            if min_dist is None and max_dist is None:
                outfile = 'eval.npz'
            else:
                outfile = 'eval_%s_%s.npz' % (min_dist, max_dist)

        # resolve case where a condition name was passed to cluster_pattern
        if cluster_pattern in self.loop_patterns.keys():
            cluster_pattern = self.loop_patterns[cluster_pattern]

        # make y_true and pvalues/qvalues (if necessary) one chrom at a time
        y_true = []
        pvalues = []
        qvalues = []
        for chrom in self.chroms:
            # load data
            disp_idx = self.load_data('disp_idx', chrom)
            loop_idx = self.load_data('loop_idx', chrom)
            row = self.load_data('row', chrom, idx=(disp_idx, loop_idx))
            col = self.load_data('col', chrom, idx=(disp_idx, loop_idx))
            clusters = load_clusters(cluster_pattern.replace('<chrom>', chrom))
            labels = np.loadtxt(label_pattern.replace('<chrom>', chrom),
                                dtype='|S7')

            # construct dist_idx
            dist = col - row
            dist_idx = np.ones(len(dist), dtype=bool)
            if min_dist is not None:
                dist_idx[dist < min_dist] = False
            if max_dist is not None:
                dist_idx[dist > max_dist] = False

            # append to y_true and pvalues/qvalues (if necessary)
            y_true.append(make_y_true(
                row[dist_idx], col[dist_idx], clusters, labels))
            if min_dist is not None or max_dist is not None:
                if rerun_bh:
                    pvalues.append(self.load_data('pvalues', chrom,
                                   idx=(loop_idx, dist_idx)))
                else:
                    qvalues.append(self.load_data('qvalues', chrom,
                                   idx=dist_idx))

        # concatenate y_true and make or load qvalues
        y_true = np.concatenate(y_true)
        if pvalues:
            qvalues = adjust_pvalues(np.concatenate(pvalues))
        elif qvalues:
            qvalues = np.concatenate(qvalues)
        else:
            qvalues, _ = self.load_data('qvalues', 'all')

        # evaluate and save to disk
        fdr, fpr, tpr, thresh = evaluate(y_true, qvalues)

        # save to disk
        np.savez('%s/%s' % (self.outdir, outfile),
                 **{'fdr': fdr, 'fpr': fpr, 'tpr': tpr, 'thresh': thresh})



import numpy as np
import scipy.sparse as sparse

from hic3defdr.util.clusters import find_clusters


def classify(row, col, value, clusters):
    """

    Parameters
    ----------
    row, col : np.ndarray
        Row and column indices, respectively, identifying pixels which
        correspond to rows of ``value``.
    value : np.ndarray
        Value to use to assign classes. Rows correspond to pixels, columns to
        conditions.
    clusters : list of set of tuple
        Only pixels in these clusters will be assigned classes.

    Returns
    -------
    list of list of set of tuple
        The outer list represents classes (one per condition). Its elements
        represent clusters of points with that class/condition.
    """
    # convert clusters to idx, parallel to row/col
    pixels = set().union(*(c for c in clusters))
    idx = np.zeros_like(row, dtype=bool)
    for i, p in enumerate(zip(row, col)):
        if p in pixels:
            idx[i] = True

    # assign classes
    classes = np.argmax(value[idx, :], axis=1)

    # re-cluster each class
    n = max(row.max(), col.max()) + 1
    class_clusters = []
    for c in range(value.shape[1]):
        coo = sparse.coo_matrix(
            (np.ones((classes == c).sum()),
             (row[idx][classes == c], col[idx][classes == c])), shape=(n, n))
        class_clusters.append(find_clusters(coo))

    return class_clusters



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from lib5c.util.plotting import plotter


@plotter
def plot_distance_bias(ob, bins, bin_labels=None, idx='disp', threshold=0.05,
                       colors=None, labels=None, xlabel='distance range',
                       legend_label='group', **kwargs):
    """
    Plots a bar plot illustrating the degree to which p-values are biased among
    different distance scales.

    This method visualizes distance bias by computing a specified percentile of
    all p-values called in an analysis, and then computing the proportion of
    pixels with p-values below this percentile in each distance bin.

    Parameters
    ----------
    ob : hic3defdr.HiC3DeFDR object or list of hic3defdr.HiC3DeFDR objects
        The analyis or analyses to inspect for distance bias.
    bins : list of tuple of int
        Each tuple represents a distance bin as a (min, max) pair, where min and
        max are distances in bin units and the ranges are inclusive. If either
        min or max is None, the distance bin will be considered unbounded on
        that end.
    bin_labels : list of str
        Pass a list of labels to describe the distance bins.
    idx : {'disp', 'loop'}
        Pass 'disp' to use p-values for all points for which dispersion was
        estimated. Pass 'loop' to only use p-values for points which are in
        loops.
    threshold : float
        The percentile to use for the comparison.
    colors : str or list of str, optional
        If ``ob`` is a single object, pass a single color to color the bars in
        the barplot. If ``ob`` is a list of objects, pass a list of colors. Pass
        None to use automatic colors.
    labels : list of str, optional
        If ``ob`` is a list of objects, you must pass a list of strings to label
        the objects. Otherwise, this kwarg does nothing.
    xlabel : str
        The label for the x-axis.
    legend_label : str
        If ``ob`` is a list of objects, the label to use for the legend title.
        Otherwise, this kwarg does nothing.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis
        The axis plotted on.
    """
    # promote ob to list and resolve labels
    if type(ob) not in [list, tuple]:
        ob = [ob]
        hue = None
        color = 'k' if colors is None else colors
        colors = None
        if labels is None:
            labels = ['group1']
    else:
        hue = legend_label
        color = None
        if labels is None:
            raise ValueError('must pass labels if ob is a list or tuple')

    # resolve bin_labels
    if bin_labels is None:
        bin_labels = []
        for min_dist, max_dist in bins:
            if min_dist is None and max_dist is not None:
                label = '<= %s' % max_dist
            elif min_dist is not None and max_dist is None:
                label = '>= %s' % min_dist
            elif min_dist is None and max_dist is None:
                label = 'all'
            else:
                label = '%s to %s' % (min_dist, max_dist)
            bin_labels.append(label)

    data = []
    for o, label in zip(ob, labels):
        # load data
        disp_idx, _ = o.load_data('disp_idx', 'all')
        if idx == 'loop':
            loop_idx, _ = o.load_data('loop_idx', 'all')
            rc_idx = (disp_idx, loop_idx)
            p_idx = loop_idx
        else:
            rc_idx = disp_idx
            p_idx = None
        disp_idx, _ = o.load_data('disp_idx', 'all')
        loop_idx, _ = o.load_data('loop_idx', 'all')
        row, _ = o.load_data('row', 'all', idx=rc_idx)
        col, _ = o.load_data('col', 'all', idx=rc_idx)
        dist = col - row
        pvalues, _ = o.load_data('pvalues', 'all', idx=p_idx)
        p_star = np.percentile(pvalues, 100*threshold)

        # process each distance bin
        for bin_label, (min_dist, max_dist) in zip(bin_labels, bins):
            dist_idx = np.ones(len(dist), dtype=bool)
            if min_dist is not None:
                dist_idx[dist < min_dist] = False
            if max_dist is not None:
                dist_idx[dist > max_dist] = False
            perc = np.mean(pvalues[dist_idx] < p_star)
            data.append({legend_label: label, xlabel: bin_label,
                         'percentage significant': perc})
    df = pd.DataFrame(data)
    sns.barplot(data=df, x=xlabel, y='percentage significant', hue=hue,
                color=color, palette=colors)
    xlim = plt.xlim()
    plt.hlines(threshold, *xlim, color='gray', linestyle='--')
    plt.xlim(xlim)



import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from lib5c.util.lowess import lowess

from hic3defdr.util.logging import eprint


def lowess_fit(x, y, logx=False, logy=False, left_boundary=None,
               right_boundary=None, frac=0.3, delta=0.01):
    """
    Opinionated convenience wrapper for lowess smoothing.

    Parameters
    ----------
    x, y : np.ndarray
        The x and y values to fit, respectively.
    logx, logy : bool
        Pass True to perform the fit on the scale of ``log(x)`` and/or
        ``log(y)``, respectively.
    left_boundary, right_boundary : float, optional
        Allows specifying boundaries for the fit, in the original ``x`` space.
        If a float is passed, the returned fit will return the farthest left or
        farthest right lowess-estimated ``y_hat`` (from the original fitting
        set) for all points which are left or right of the specified left or
        right boundary point, respectively. Pass None to use linear
        extrapolation for these points instead.
    frac : float
        The lowess smoothing fraction to use.
    delta : float
        Distance (on the scale of ``x`` or ``log(x)``) within which to use
        linear interpolation when constructing the initial fit, expressed as a
        fraction of the range of ``x`` or ``log(x)``.

    Returns
    -------
    function
        This function takes in ``x`` values on the original ``x`` scale and
        returns estimated ``y`` values on the original ``y`` scale (regardless
        of what is passed for ``logx`` and ``logy``). This function will still
        return sane estimates for ``y`` even at points not in the original
        fitting set by performing linear interpolation in the space the fit was
        performed in.

    Notes
    -----
    No filtering of input values is performed; clients are expected to handle
    this if desired. NaN values should not break the function, but ``x`` points
    with zero values passed when ``logx`` is True are expected to break the
    function.

    The default value of the ``delta`` parameter is set to be non-zero, matching
    the behavior of lowess smoothing in R and improving performance.

    Linear interpolation between x-values in the original fitting set is used to
    provide a familiar functional interface to the fitted function.

    Boundary conditions on the fitted function are exposed via ``left_boundary``
    and ``right_boundary``, mostly as a convenience for points where ``x == 0``
    when fitting was performed on the scale of ``log(x)``.

    When ``left_boundary`` or ``right_boundary`` are None (this is the default)
    the fitted function will be linearly extrapolated for points beyond the
    lowest and highest x-values in ``x``.
    """
    if logx:
        x = np.log(x)
    if logy:
        y = np.log(y)

    res = lowess(y, x, frac=frac, delta=(np.nanmax(x) - np.nanmin(x)) * delta)
    sorted_x = res[:, 0]
    sorted_y_hat = res[:, 1]

    def fit(x_star):
        if logx:
            new_x = np.log(x_star)
        else:
            new_x = x_star
        _, idx = np.unique(sorted_x, return_index=True)
        y_hat = interp1d(sorted_x[idx], sorted_y_hat[idx],
                         fill_value='extrapolate', assume_sorted=True)(new_x)
        if left_boundary is not None:
            y_hat[x_star <= left_boundary] = sorted_y_hat[0]
        if right_boundary is not None:
            y_hat[x_star >= right_boundary] = sorted_y_hat[-1]
        if logy:
            y_hat = np.exp(y_hat)
        return y_hat

    return fit


def weighted_lowess_fit(x, y, logx=False, logy=False, left_boundary=None,
                        right_boundary=None, frac=None, auto_frac_factor=15.,
                        delta=0.01, w=20, power=1./4,
                        interpolate_before_increase=True):
    """
    Performs lowess fitting as in ``lowess_fit()``, but weighting the data
    points automatically according to the precision in the ``y`` values as
    estimated by a rolling window sample variance.

    Points are weighted proportionally to a specified power ``power`` of their
    precision by adding duplicated points to the dataset. This should
    approximate the effects of a true weighted lowess fit, with the caveat that
    the weights are rounded a bit.

    Weighting the data points according to this rolling window sample variance
    is probably only a good idea if the marginal distribution of ``x`` values is
    uniform.

    Parameters
    ----------
    x, y : np.ndarray
        The x and y values to fit, respectively.
    logx, logy : bool
        Pass True to perform the fit on the scale of ``log(x)`` and/or
        ``log(y)``, respectively.
    left_boundary, right_boundary : float, optional
        Allows specifying boundaries for the fit, in the original ``x`` space.
        If a float is passed, the returned fit will return the farthest left or
        farthest right lowess-estimated ``y_hat`` (from the original fitting
        set) for all points which are left or right of the specified left or
        right boundary point, respectively. Pass None to use linear
        extrapolation for these points instead.
    frac : float, optional
        The lowess smoothing fraction to use. Pass None to use the default:
        ``auto_frac_factor`` divided by the product of the average of the
        unscaled weights and the largest scaled weight.
    auto_frac_factor : float
        When ``frac`` is None, this factor scales the automatically determined
        fraction parameter.
    delta : float
        Distance (on the scale of ``x`` or ``log(x)``) within which to use
        linear interpolation when constructing the initial fit, expressed as a
        fraction of the range of ``x`` or ``log(x)``.
    w : int
        The size of the rolling window to use when estimating the precision of
        the y values.
    power : float
        Precisions will be taken to this power to obtain unscaled weights.
    interpolate_before_increase : bool
        Hacky flag introduced to handle quirk of Hi-C dispersion vs distance
        relationships in which dispersion is elevated at extremely short
        distances. When True, this function will identify a group of points with
        the lowest x-values across which the y-value is monotonically
        decreasing. These points will be included in the variance estimation,
        but will be excluded from lowess fitting. Linear interpolation will be
        used at these x-values instead, since it is hard to convince lowess to
        follow a sharp change in the trend that is only supported by 3-4 data
        points out of 200-500 total data points, even with our best attempts at
        weighting. Pass False to perform a simple weighted lowess fit with no
        linear interpolation.

    Returns
    -------
    function
        This function takes in ``x`` values on the original ``x`` scale and
        returns estimated ``y`` values on the original ``y`` scale (regardless
        of what is passed for ``logx`` and ``logy``). This function will still
        return sane estimates for ``y`` even at points not in the original
        fitting set by performing linear interpolation in the space the fit was
        performed in.
    """
    n = len(y)
    i = np.arange(n)
    sort_idx = np.argsort(x)
    x = x[sort_idx].copy()
    y = y[sort_idx].copy()

    # compute rolling var
    var = pd.Series(y).rolling(window=w, center=True).var().values

    # convert to precision
    prec = 1 / var

    # apply power to obtain weights
    weight = np.power(prec, power)

    # scale to make smallest weight 1
    min_weight = np.nanmin(weight)
    scaled_weight = weight * (1 / min_weight)

    # clip "infinite weight" points to the max weight
    max_weight = np.nanmax(scaled_weight)
    scaled_weight[np.isinf(scaled_weight)] = max_weight

    # fill left and right side nan's
    # get the first non-nan precision (for filling left side)
    left_weight = scaled_weight[np.argmax(np.isfinite(scaled_weight))]
    left_fill_idx = np.isnan(scaled_weight) & (i < n/2)
    right_fill_idx = np.isnan(scaled_weight) & (i > n/2)
    scaled_weight[left_fill_idx] = left_weight
    # fill right side with 1s
    scaled_weight[right_fill_idx] = 1
    assert np.all(np.isfinite(scaled_weight))

    # floor and convert to int
    floored_weight = np.floor(scaled_weight).astype(int)

    # find index of first increase
    inc_idx = np.argmax(np.diff(y) > 0) + 1 if interpolate_before_increase \
        else 0

    # create duplicated data
    expanded_xs = []
    expanded_ys = []
    for i in range(inc_idx, n):
        m = floored_weight[i]
        if not np.isfinite(m):
            continue
        expanded_xs.extend([x[i]] * m)
        expanded_ys.extend([y[i]] * m)

    # resolve frac
    if frac is None:
        frac_auto = auto_frac_factor / (max_weight * np.nanmean(weight))
        frac = max(min(frac_auto, 2./3), 0.05)
        eprint('  using auto-determined lowess fraction of %.3f' % frac)

    # lowess fit duplicated data
    lowess_fn = lowess_fit(
        np.array(expanded_xs), np.array(expanded_ys), logx=logx, logy=logy,
        left_boundary=left_boundary, right_boundary=right_boundary, frac=frac,
        delta=delta)

    def fit(x_star):
        interp_y_hat = interp1d(x, y, bounds_error=False,
                                fill_value=(y[0], 'extrapolate'))(x_star)
        fit_y_hat = lowess_fn(x_star)
        interp_idx = x_star < x[inc_idx]
        if type(interp_idx) == bool:
            if interp_idx:
                return interp_y_hat
            else:
                return fit_y_hat
        fit_y_hat[interp_idx] = interp_y_hat[interp_idx]
        return fit_y_hat

    return fit



from multiprocessing import Pool, cpu_count

import dill

from hic3defdr.util.progress import tqdm_avail, tqdm


def _pack_for_apply(fn, kwargs_list):
    return [dill.dumps((fn, kwargs)) for kwargs in kwargs_list]


def _unpack_for_apply(payload):
    fn, kwargs = dill.loads(payload)
    return fn(**kwargs)


def parallel_apply(fn, kwargs_list, n_threads=None):
    """
    Applies a function in parallel over a list of kwarg dicts, using a
    specified number of threads.

    Parameters
    ----------
    fn : function
        The function to parallelize.
    kwargs_list : list of dict
        The function will be called ``len(kwargs_list)`` times. Each time it is
        called it will use a different element of this list to determine the
        kwargs the function will be called with.
    n_threads : int
        Specify the number of subprocesses to use for parallelization. Pass -1
        to use as many subprocesses as there are CPUs.
    """
    if n_threads == -1:
        n_threads = cpu_count()
    pool = Pool(n_threads)
    pool.map(_unpack_for_apply, _pack_for_apply(fn, kwargs_list))
    pool.close()
    pool.join()


def _pack_for_map(fn, kwargs):
    return dill.dumps((fn, kwargs))


def _unpack_for_map(payload):
    fn, kwargs = dill.loads(payload)
    idx = kwargs['_idx']
    del kwargs['_idx']
    return idx, fn(**kwargs)


def parallel_map(fn, kwargs_list, n_threads=None):
    """
    Maps a function in parallel over a list of kwarg dicts, using a specified
    number of threads.

    Parameters
    ----------
    fn : function
        The function to parallelize.
    kwargs_list : list of dict
        The function will be called ``len(kwargs_list)`` times. Each time it is
        called it will use a different element of this list to determine the
        kwargs the function will be called with.
    n_threads : int
        Specify the number of subprocesses to use for parallelization. Pass -1
        to use as many subprocesses as there are CPUs.
    """
    # resolve n_threads
    if n_threads == -1:
        n_threads = cpu_count()

    # create a progress bar and result list
    n = len(kwargs_list)
    pbar = tqdm(total=n) if tqdm_avail else None
    result = [None] * n

    # fill in magic kwarg _idx
    for i in range(n):
        kwargs_list[i]['_idx'] = i

    # callback for apply_async
    # payload is the return value of _unpack_for_map
    def update(payload):
        i, r = payload
        result[i] = r
        if tqdm_avail:
            pbar.update()

    # process in parallel
    pool = Pool(n_threads)
    for kwargs in kwargs_list:
        pool.apply_async(_unpack_for_map, args=(_pack_for_map(fn, kwargs),),
                         callback=update)
    pool.close()
    pool.join()
    if pbar is not None:
        pbar.close()
    return result



import pandas as pd
import dill as pickle

from lib5c.util.system import check_outdir

from hic3defdr.analysis.core import CoreHiC3DeFDR
from hic3defdr.analysis.analysis import AnalyzingHiC3DeFDR
from hic3defdr.analysis.simulation import SimulatingHiC3DeFDR
from hic3defdr.analysis.plotting import PlottingHiC3DeFDR


class HiC3DeFDR(CoreHiC3DeFDR, AnalyzingHiC3DeFDR, SimulatingHiC3DeFDR,
                PlottingHiC3DeFDR):
    """
    Main object for hic3defdr analysis.

    Attributes
    ----------
    raw_npz_patterns : list of str
        File path patterns to ``scipy.sparse`` formatted NPZ files containing
        raw contact matrices for each replicate, in order. Each file path
        pattern should contain at least one '<chrom>' which will be replaced
        with the chromosome name when loading data for specific chromosomes.
    bias_patterns : list of str
        File path patterns to ``np.savetxt()`` formatted files containing bias
        vector information for each replicate, in order. ach file path pattern
        should contain at least one '<chrom>' which will be replaced with the
        chromosome name when loading data for specific chromosomes.
    chroms : list of str
        List of chromosome names as strings. These names will be substituted in
        for '<chroms>' in the ``raw_npz_patterns`` and ``bias_patterns``.
    design : pd.DataFrame or str
        Pass a DataFrame with boolean dtype whose rows correspond to replicates
        and whose columns correspond to conditions. Replicate and condition
        names will be inferred from the row and column labels, respectively. If
        you pass a string, the DataFrame will be loaded via
        ``pd.read_csv(design, index_col=0)``.
    outdir : str
        Specify a directory to store the results of the analysis. Two different
        HiC3DeFDR analyses cannot co-exist in the same directory. The directory
        will be created if it does not exist.
    dist_thresh_min, dist_thresh_max : int
        The minimum and maximum interaction distance (in bin units) to include
        in the analysis.
    bias_thresh : float
        Bins with a bias factor below this threshold or above its reciprocal in
        any replicate will be filtered out of the analysis.
    mean_thresh : float
        Pixels with mean value below this threshold will be filtered out at the
        dispersion fitting stage.
    loop_patterns : dict of str
        Keys should be condition names as strings, values should be file path
        patterns to sparse JSON formatted cluster files representing called
        loops in that condition. Each file path pattern should contain at least
        one '<chrom>' which will be replaced with the chromosome name when
        loading data for specific chromosomes.
    """
    def __init__(self, raw_npz_patterns, bias_patterns, chroms, design, outdir,
                 dist_thresh_min=4, dist_thresh_max=200, bias_thresh=0.1,
                 mean_thresh=1.0, loop_patterns=None):
        """
        Base constructor. See ``help(HiC3DeFDR)`` for details.
        """
        self.raw_npz_patterns = raw_npz_patterns
        self.bias_patterns = bias_patterns
        self.chroms = chroms
        if type(design) == str:
            self.design = pd.read_csv(design, index_col=0)
        else:
            self.design = design
        self.outdir = outdir
        self.dist_thresh_min = dist_thresh_min
        self.dist_thresh_max = dist_thresh_max
        self.bias_thresh = bias_thresh
        self.mean_thresh = mean_thresh
        self.loop_patterns = loop_patterns
        state = self.__dict__.copy()
        del state['outdir']
        check_outdir(self.picklefile)
        with open(self.picklefile, 'wb') as handle:
            pickle.dump(state, handle, -1)



from hic3defdr import HiC3DeFDR

repnames = ['ES_1', 'ES_3', 'NPC_2', 'NPC_4']
#chroms = ['chr%i' % i for i in range(1, 20)] + ['chrX']
#chroms = ['chr19']
#chroms = ['chr3']
chroms = ['chr18']
base_path = 'c:/lab-data/bonev/'
f = HiC3DeFDR(
    [base_path + '<rep>/<chrom>_raw.npz'.replace('<rep>', repname)
     for repname in repnames],
    [base_path + '<rep>/<chrom>_kr.bias'.replace('<rep>', repname)
     for repname in repnames],
    chroms, 'design.csv', 'output-test',
    loop_patterns={c: base_path + 'clusters_0.025/%s_<chrom>_clusters.json' % c
                   for c in ['ES', 'NPC']}
)
f = HiC3DeFDR.load('output-test')
#f.run_to_qvalues(norm='median_of_ratios')
#f.run_to_qvalues()
f.run_to_qvalues(n_bins_norm=None)
#f.prepare_data()
#f.estimate_disp()
#f.lrt()
#f.bh()
f.plot_dd_curves('chr18', outfile='dd2.png')
f.plot_ma('chr18', legend=True, outfile='ma_legend2.png')
#f.plot_grid('chr18', 2218, 2236, 50, outfile='grid.png')
#f.plot_dispersion_fit('chr18', 'ES', yaxis='disp', outfile='disp.png')
#f.plot_pvalue_distribution(outfile='pvalue_dist.png')



import numpy as np
import matplotlib.pyplot as plt

from lib5c.util.plotting import plotter


@plotter
def plot_pvalue_histogram(data, xlabel='pvalue', **kwargs):
    """
    Plots a p-value or q-value distribution.

    Parameters
    ----------
    data : np.ndarray
        The p-values or q-values to plot.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis
        The axis plotted on.
    """
    plt.hist(data, bins=np.linspace(0, 1, 21))
    plt.ylabel('number of pixels')



import numpy as np

from hic3defdr.util.progress import tqdm_maybe as tqdm

try:
    from sklearn.metrics import roc_curve, confusion_matrix

    sklearn_avail = True
except ImportError:
    sklearn_avail = False
    roc_curve = None
    confusion_matrix = None


def make_y_true(row, col, clusters, labels):
    """
    Makes a boolean vector of the true labels for each pixel, given a list of
    clusters and the true label for each cluster.

    Parameters
    ----------
    row, col : np.ndarray
        The row and column indices of pixels to be labeled.
    clusters : list of list of tuple
        The outer list is a list of clusters. Each cluster is a list of (i, j)
        tuples marking the position of significant points which belong to that
        cluster.
    labels : list of str
        List of labels for each cluster, parallel to ``clusters``.

    Returns
    -------
    np.ndarray
        Boolean vector with the same length as ``row``/``col``. It's `i`th
        element is False when the pixel at `(row[i], col[i])` is in a cluster
        with label 'constit' and is True otherwise.
    """
    sig_idx = ~(labels == 'constit')
    sig_pixels = set().union(*[c for i, c in enumerate(clusters) if sig_idx[i]])
    return np.array([True if (r, c) in sig_pixels else False
                     for r, c in zip(row, col)])


def evaluate(y_true, qvalues, n_fdr_points=100):
    """
    Evaluates how good a vector of q-values (or p-values) is at predicting the
    vector of true labels.

    Parameters
    ----------
    y_true : np.ndarray
        The boolean vector of true labels.
    qvalues : np.ndarray
        Vector of q-values or p-values which are supposed to predict the boolean
        label in ``y_true``.
    n_fdr_points : int
        The maximum number of points at which to compute FDR. The FDR
        computation is not parallelized so increasing this number will slow down
        the evaluation. The default value of 100 should be sufficient to
        visualize the FDR control curve.

    Returns
    -------
    fdr, fpr, tpr, thresh : np.ndarray
        Parallel arrays of the FDR, FPR, TPR, and thresholds (in ``1 - qvalue``)
        space which specify the FDR, FPR, and and TPR at each threshold. The
        thresholds are selected to represent the convex edge of the ROC curve.
        The FDR will only be evaluated at about 100 selected thresholds and
        will be set to ``np.nan`` at the un-evaluated thresholds.
    """
    if not sklearn_avail:
        raise ImportError('failed to import scikit-learn - is it installed?')
    y_pred = 1 - qvalues
    fpr, tpr, thresh = roc_curve(y_true, y_pred)
    fdr = np.ones_like(fpr) * np.nan
    for i in tqdm(range(np.argmax(tpr > 0), len(thresh),
                        max(len(thresh)/n_fdr_points, 1))):
        fdr[i] = compute_fdr(y_true, y_pred >= thresh[i])
    return fdr, fpr, tpr, thresh


def compute_fdr(y_true, y_pred):
    """
    Computes the observed false discovery rate from boolean vectors of true and
    predicted labels.

    Parameters
    ----------
    y_true, y_pred : np.ndarray
        Boolean vectors of the true and predicted labels, respectively.

    Returns
    -------
    float
        The false discovery rate.
    """
    if not sklearn_avail:
        raise ImportError('failed to import scikit-learn - is it installed?')
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
    return fp / float(fp + tp)



import numpy as np


def equal_bin(data, n_bins):
    """
    Bins ``data`` into ``n_bins`` bins, with an equal number of points in each
    bin.

    https://stackoverflow.com/a/40895507

    Parameters
    ----------
    data : np.ndarray
        The data to bin.
    n_bins : int
        The number of bins to bin into.

    Returns
    -------
    np.ndarray
        A vector of integers representing the bin index for each entry in
        ``data``.
    """
    idx = np.linspace(0, n_bins, data.size+0.5, endpoint=0).astype(int)
    return idx[data.argsort().argsort()]



import numpy as np
from scipy.optimize import minimize_scalar
from scipy.special import gammaln

from hic3defdr.util.scaled_nb import inverse_mvr, equalize
from hic3defdr.util.binning import equal_bin
from hic3defdr.util.lowess import lowess_fit


def qcml(data, f=None, max_iter=10, tol=1e-4):
    """
    Estimates the dispersion parameter of a NB distribution given the data via
    quantile-adjusted conditional maximum likelihood.

    One common dispersion will be estimated across all pixels in the data,
    though the individual pixels in the data may have distinct means.

    Parameters
    ----------
    data : np.ndarray
        Rows should correspond to pixels, columns to replicates.
    f : np.ndarray
        Matrix of scaling factors parallel to ``data``. Pass None to skip
        scaling.

    Returns
    -------
    float
        The estimated dispersion.
    """
    if f is None:
        f = np.ones_like(data, dtype=float)
    disp = 0.01
    delta = np.inf
    it = 1
    while delta > tol and it <= max_iter:
        pseudodata = equalize(data, f, disp)
        new_disp = cml(pseudodata)
        delta = np.abs(disp - new_disp)
        disp = new_disp
        if delta < tol:
            break
    return disp


def cml(data, f=None):
    """
    Estimates the dispersion parameter of a NB distribution given the data via
    conditional maximum likelihood.

    One common dispersion will be estimated across all pixels in the data,
    though the individual pixels in the data may have distinct means.

    Parameters
    ----------
    data : np.ndarray
        Rows should correspond to pixels, columns to replicates.
    f : np.ndarray
        Matrix of scaling factors parallel to ``data``. Pass None to skip
        scaling.

    Returns
    -------
    float
        The estimated dispersion.
    """
    if f is not None:
        data /= f
    n = data.shape[1]
    z = np.sum(data, axis=1)

    def nll(delta):
        r = 1./delta - 1
        return -np.sum((np.sum(gammaln(data + r), axis=1) + gammaln(n*r) -
                        gammaln(z+n*r) - n * gammaln(r)))

    res = minimize_scalar(nll, bounds=(1e-4, 100./(100+1)), method='bounded')
    assert res.success
    delta_hat = res.x
    return delta_hat / (1-delta_hat)


def mme_per_pixel(data, f=None):
    """
    Estimates the dispersion parameter of a separate NB distribution for each
    pixel given the data using a method of moments approach.

    Parameters
    ----------
    data : np.ndarray
        Rows should correspond to pixels, columns to replicates.
    f : np.ndarray
        Matrix of scaling factors parallel to ``data``. Pass None to skip
        scaling.

    Returns
    -------
    np.ndarray
        The estimated dispersion for each pixel.
    """
    if f is not None:
        data /= f
    m = np.mean(data, axis=1)
    v = np.var(data, axis=1, ddof=1)
    return inverse_mvr(m, v)


def mme(data, f=None):
    """
    Estimates the dispersion parameter of a NB distribution given the data using
    a method of moments approach.

    One common dispersion will be estimated across all pixels in the data,
    though the individual pixels in the data may have distinct means.

    Parameters
    ----------
    data : np.ndarray
        Rows should correspond to pixels, columns to replicates.
    f : np.ndarray
        Matrix of scaling factors parallel to ``data``. Pass None to skip
        scaling.

    Returns
    -------
    float
        The estimated dispersion.
    """
    if f is not None:
        data /= f
    return np.nanmean(mme_per_pixel(data))


def estimate_dispersion(data, cov, estimator='qcml', n_bins=100, logx=False):
    """
    Estimates trended dispersion for each point in ``data`` with respect to a
    covariate ``cov``, using ``estimator`` to estimate the dispersion within
    each of ``n_bins`` equal-number bins and fitting a curve through the per-bin
    dispersion estimates using lowess smoothing.

    This function is deprecated and has no callers.

    Parameters
    ----------
    data : np.ndarray
        Rows should correspond to pixels, columns to replicates.
    cov : np.ndarray
        A vector of covariate values per pixel.
    estimator : 'cml', 'qcml', 'mme', or a function
        Pass 'cml', 'qcml', 'mme' to use conditional maximum likelihood (CML),
        qnorm-CML (qCML), or method of moments estimation (MME) to estimate the
        dispersion within each bin. Pass a function that takes in a (pixels,
        replicates) shaped array of data and returns a dispersion value to use
        that instead.
    n_bins : int
        The number of bins to use when binning the pixels according to ``cov``.
    logx : bool
        Whether or not to perform the lowess fit in log x space.

    Returns
    -------
    smoothed_disp : np.ndarray
        Vector of smoothed dispersion estimates for each pixel.
    cov_per_bin, disp_per_bin : np.ndarray
        Average covariate value and estimated dispersion value, respectively,
        per bin.
    disp_smooth_func : function
        Vectorized function that returns a smoothed dispersion given the value
        of the covariate.
    """
    if type(estimator) == str and estimator not in {'cml', 'qcml', 'mme'}:
        raise ValueError('estimator must be cml, qcml, mme, or a function')
    disp_func = globals()[estimator] if estimator in globals().keys() \
        else estimator
    bins = equal_bin(cov, n_bins)
    cov_per_bin = np.array([np.mean(cov[bins == b]) for b in range(n_bins)])
    disp_per_bin = np.array([disp_func(data[bins == b, :])
                             for b in range(n_bins)])
    cov_idx = cov_per_bin > 0
    disp_smooth_func = lowess_fit(cov_per_bin[cov_idx], disp_per_bin[cov_idx],
                                  left_boundary=None, logx=logx, logy=True)
    smoothed_disp = disp_smooth_func(cov)
    return smoothed_disp, cov_per_bin, disp_per_bin, disp_smooth_func



import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from lib5c.util.plotting import plotter
from lib5c.plotters.colormaps import get_colormap

from hic3defdr.util.matrices import dilate
from hic3defdr.util.thresholding import threshold_and_cluster, size_filter
from hic3defdr.util.clusters import clusters_to_coo
from hic3defdr.util.classification import classify
from hic3defdr.plotting.heatmap import plot_heatmap


@plotter
def plot_grid(i, j, w, row, col, raw, scaled, mu_hat_alt, mu_hat_null, qvalues,
              disp_idx, loop_idx, design, fdr, cluster_size, vmax=100,
              fdr_vmid=0.05,
              color_cycle=('blue', 'green', 'purple', 'yellow', 'cyan', 'red'),
              despine=False, **kwargs):
    """
    Plots a combination visualization grid focusing on a specific pixel on a
    specific chromosome, combining heatmaps, cluster outlines, and stripplots.

    Parameters
    ----------
    i, j : int
        The row and column index of the pixel to focus on.
    w : int
        The size of the heatmap will be ``2*w + 1`` bins in each dimension.
    row, col : np.ndarray
        The row and column indices corresponding to the rows of the ``raw`` and
        ``scaled`` matrices.
    raw, scaled : np.ndarray
        The raw and scaled data for each pixel (rows) and each replicate
        (columns).
    mu_hat_alt, mu_hat_null : np.ndarray
        The estimated mean parameter under the alternate and null model,
        respectively. First dimension is pixels for which dispersion was
        estimated, whose row and column coordinates in the complete square
        matrix are given by ``row[disp_idx]`` and ``col[disp_idx]``,
        respectively. Columns of ``mu_hat_alt`` correspond to conditions, while
        ``mu_hat_null`` has no second dimension.
    qvalues : np.ndarray
        Vector of q-values called per pixel whose dispersion was estimated and
        which lies in a loop. The row and column coordinates in the complete
        square matrix are given by ``row[disp_idx][loop_idx]`` and
        ``col[disp_idx][loop_idx]``, respectively.
    disp_idx : np.ndarray
        Boolean matrix indicating which pixels in ``zip(row, col)`` had their
        dispersion estimated.
    loop_idx : np.ndarray
        Boolean matrix indicating which pixels in
        ``zip(row[disp_idx], col[disp_idx])`` lie within loops.
    design : pd.DataFrame
        Pass a DataFrame with boolean dtype whose rows correspond to replicates
        and whose columns correspond to conditions. Replicate and condition
        names will be inferred from the row and column labels, respectively.
    fdr : float
        The FDR threshold to use when outlining clusters.
    cluster_size : int
        The cluster size threshold to use when outlining clusters.
    vmax : float
        The maximum of the colorscale to use when plotting normalized
        heatmaps.
    fdr_vmid : float
        The FDR value at the middle of the colorscale used for plotting the
        q-value heatmap.
    color_cycle : list of matplotlib colors
        The color cycle to use over conditions.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis, grid of pyplot axes, function
        The first pyplot axis returned is injected by ``@plotter``. The grid of
        pyplot axes is the second return value from the call to
        ``plt.subplots()`` that is used to create the grid. The function takes
        two args, an FDR and a cluster size, and redraws the cluster outlines
        using the new parameters.
    """
    # precompute some things
    max_reps = np.max(np.sum(design, axis=0))
    idx = np.where((row[disp_idx] == i) & (col[disp_idx] == j))[0][0]
    extent = [-0.5, 2 * w + 0.5, -0.5, 2 * w + 0.5]
    rs, cs = slice(i - w, i + w + 1), slice(j - w, j + w + 1)
    f = raw[disp_idx] / scaled[disp_idx]
    n = max(row.max(), col.max()) + 1
    mu_hat_alt = np.dot(mu_hat_alt, design.values.T)
    mean = np.dot(scaled, design) / np.sum(design, axis=0).values

    # set the color cycle over conditions
    cc = color_cycle

    # plot
    fig, ax = plt.subplots(design.shape[1] + 1, max_reps + 1,
                           figsize=(design.shape[1] * 6, max_reps * 6))
    bwr = get_colormap('bwr', set_bad='lightgray')
    red = get_colormap('Reds', set_bad='lightgray')
    plot_heatmap(
        row[disp_idx][loop_idx], col[disp_idx][loop_idx], -np.log10(qvalues),
        rs, cs, cmap=bwr, vmax=-np.log10(fdr_vmid)*2, ax=ax[-1, 0])
    ax[-1, 0].set_title('qvalues')
    for c in range(design.shape[1]):
        plot_heatmap(
            row, col, mean[:, c], rs, cs, cmap=red, vmax=vmax, ax=ax[c, 0])
        ax[c, 0].set_ylabel(design.columns[c])
        ax_idx = 1
        for r in range(design.shape[0]):
            if not design.values[r, c]:
                continue
            plot_heatmap(
                row, col, scaled[:, r], rs, cs, cmap=red, vmax=vmax,
                ax=ax[c, ax_idx])
            ax[c, ax_idx].set_title(design.index[r])
            ax_idx += 1
    ax[0, 0].set_title('mean')
    for r in range(design.shape[1] + 1):
        for c in range(max_reps + 1):
            if r < design.shape[1] - 1:
                ax[r, c].get_xaxis().set_ticks([])
            if c > 0:
                ax[r, c].get_yaxis().set_ticks([])
            if r == design.shape[1] and c == 0:
                break

    sns.stripplot(
        data=[scaled[disp_idx, :][idx, design.values[:, c]]
              for c in range(design.shape[1])],
        palette=[cc[c % len(cc)] for c in range(design.shape[1])],
        ax=ax[-1, 1]
    )
    for c in range(design.shape[1]):
        ax[-1, 1].hlines(
            mu_hat_alt[idx, np.where(design.values[:, c])[0][0]],
            c - 0.1, c + 0.1, color=cc[c % len(cc)],
            label='alt' if c == 0 else None)
        ax[-1, 1].hlines(
            mu_hat_null[idx], c - 0.1, c + 0.1, color=cc[c % len(cc)],
            linestyles='--', label='null' if c == 0 else None)
    ax[-1, 1].set_xticklabels(design.columns.tolist())
    ax[-1, 1].set_title('normalized values')
    ax[-1, 1].set_xlabel('condition')
    ax[-1, 1].legend()
    sns.despine(ax=ax[-1, 1])

    sns.stripplot(
        data=[[raw[disp_idx, :][idx, r]]
              for r in range(design.shape[0])],
        palette=[cc[c % len(cc)] for c in np.where(design)[1]],
        ax=ax[-1, 2]
    )
    for r in range(design.shape[0]):
        ax[-1, 2].hlines(
            mu_hat_alt[idx, r] * f[idx, r], r - 0.1, r + 0.1,
            color=cc[np.where(design)[1][r] % len(cc)],
            label='alt' if r == 0 else None)
        ax[-1, 2].hlines(
            mu_hat_null[idx] * f[idx, r], r - 0.1, r + 0.1,
            color=cc[np.where(design)[1][r] % len(cc)], linestyles='--',
            label='null' if r == 0 else None)
    ax[-1, 2].set_xticklabels(design.index.tolist())
    ax[-1, 2].set_title('raw values')
    ax[-1, 2].set_xlabel('replicate')
    ax[-1, 2].legend()
    sns.despine(ax=ax[-1, 2])

    contours = []
    clusters = {}

    def outline_clusters(fdr, cluster_size):
        if fdr not in clusters:
            clusters[fdr] = {}
            clusters[fdr]['base'] = threshold_and_cluster(
                qvalues, row[disp_idx][loop_idx], col[disp_idx][loop_idx], fdr)
        if cluster_size not in clusters[fdr]:
            clusters[fdr][cluster_size] = dict(zip(
                range(design.shape[1]),
                map(lambda x: clusters_to_coo(size_filter(x, cluster_size),
                                              (n, n)).tocsr()[rs, cs].toarray(),
                    classify(
                        row[disp_idx][loop_idx],
                        col[disp_idx][loop_idx],
                        mu_hat_alt[loop_idx],
                        size_filter(clusters[fdr]['base'][0], cluster_size)
                ))
            ))
            clusters[fdr][cluster_size]['insig'] = clusters_to_coo(
                size_filter(clusters[fdr]['base'][1], cluster_size),
                (n, n)).tocsr()[rs, cs].toarray()
        if contours:
            for contour in contours:
                for coll in contour.collections:
                    coll.remove()
            del contours[:]
        for c in range(design.shape[1]):
            contour = ax[-1, 0].contour(
                dilate(clusters[fdr][cluster_size][c], 2),
                [0.5], colors=cc[c % len(cc)], linewidths=3, extent=extent)
            contour.collections[0].set_label(design.columns[c])
            contours.append(contour)
        contour = ax[-1, 0].contour(
            dilate(clusters[fdr][cluster_size]['insig'], 2),
            [0.5], colors='gray', linewidths=3, extent=extent)
        contour.collections[0].set_label('constitutive')
        contours.append(contour)
        for cond_idx in range(design.shape[1]):
            for c in range(design.shape[1]):
                contour = ax[cond_idx, 0].contour(
                    dilate(clusters[fdr][cluster_size][c], 2),
                    [0.5], colors=cc[c % len(cc)], linewidths=3, extent=extent)
                contour.collections[0].set_label(design.columns[c])
                contours.append(contour)
            contour = ax[cond_idx, 0].contour(
                dilate(clusters[fdr][cluster_size]['insig'], 2),
                [0.5], colors='gray', linewidths=3, extent=extent)
            contour.collections[0].set_label('constitutive')
            contours.append(contour)

    outline_clusters(fdr, cluster_size)
    ax[-1, 0].legend(loc='lower left')

    return ax, outline_clusters



import numpy as np
import matplotlib.pyplot as plt

from lib5c.util.plotting import plotter

try:
    from sklearn.metrics import auc

    sklearn_avail = True
except ImportError:
    sklearn_avail = False
    auc = None


@plotter
def plot_roc(eval_results, labels, colors=None, **kwargs):
    """
    Plots an ROC curve from a list of ``eval.npz``-style results.

    Parameters
    ----------
    eval_results : list of dict-like
        The dicts should have keys 'thresh', 'fpr', and 'tpr' whose values are
        parallel vectors describing the thresholds, FPR, and TPR to use for the
        ROC curve. Each dict in the list represents a different ROC curve which
        will be overlayed in the plot.
    labels : list of str
        List of labels parallel to ``eval_results`` providing names for each
        curve.
    colors : list of valid matplotlib colors, optional
        Colors for each ROC curve. Pass None to automatically assign colors.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis
        The axis plotted on.
    """
    # check for sklearn
    if not sklearn_avail:
        raise ImportError('failed to import scikit-learn - is it installed?')

    # resolve colors
    if colors is None:
        colors = ['C%i' % i for i in range(len(labels))]

    for i, (res, label, color) in enumerate(zip(eval_results, labels, colors)):
        # unbox results
        fpr = res['fpr']
        tpr = res['tpr']
        thresh = res['thresh']

        # plot
        plt.plot(fpr, tpr, color=color,
                 label='%s (AUC = %0.2f)' % (label, auc(fpr, tpr)))

        # annotate thresholds if there's only one curve
        if len(labels) == 1:
            thresh_idx = np.arange(len(thresh))[slice(1, None, len(thresh)/10)]
            for t in thresh_idx:
                plt.annotate(
                    '%.2f' % (1 - thresh[t]), xy=(fpr[t], tpr[t]),
                    xytext=(fpr[t] + 0.05, tpr[t] - 0.05),
                    arrowprops=dict(facecolor='k', width=3, headwidth=6,
                                    headlength=10))

    # 45-degree line, limits, labels, legend
    plt.plot([0, 1], [0, 1], color='gray', linestyle='--')
    plt.axis('scaled')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.xlabel('FPR')
    plt.ylabel('TPR')
    plt.legend(loc='lower right')



import os

import numpy as np
import scipy.sparse as sparse

from lib5c.util.statistics import adjust_pvalues

import hic3defdr.util.scaling as scaling
import hic3defdr.util.dispersion as dispersion
from hic3defdr.util.logging import eprint
from hic3defdr.util.matrices import sparse_union
from hic3defdr.util.clusters import load_clusters, save_clusters
from hic3defdr.util.lowess import lowess_fit, weighted_lowess_fit
from hic3defdr.util.lrt import lrt
from hic3defdr.util.thresholding import threshold_and_cluster, size_filter
from hic3defdr.util.classification import classify
from hic3defdr.util.progress import tqdm_maybe as tqdm
from hic3defdr.util.parallelization import parallel_apply, parallel_map


class AnalyzingHiC3DeFDR(object):
    """
    Mixin class containing analysis functions for HiC3DeFDR.
    """
    def prepare_data(self, chrom=None, norm='conditional_mor', n_bins=-1,
                     n_threads=-1, verbose=True):
        """
        Prepares raw and normalized data for analysis.

        Parameters
        ----------
        chrom : str
            The name of the chromosome to prepare raw data for. Pass None to run
            for all chromosomes in series.
        norm : str
            The method to use to account for differences in sequencing depth.
            Valid options are:
                * simple_scaling: scale each replicate to equal total depth
                * median_of_ratios: use median of ratios normalization, ignoring
                  pixels at which any replicate has a zero
                * conditional_scaling: apply simple scaling independently at
                  each distance scale
                * conditional_mor: apply median of ratios independently at each
                  distance scale
        n_bins : int, optional
            Number of distance bins to use during scaling normalization if
            ``norm`` is one of the conditional options. Pass 0 or None to match
            pixels by exact distance. Pass -1 to use a reasonable default: 1/5
            of ``self.dist_thesh_max``.
        n_threads : int
            The number of threads (technically GIL-avoiding child processes) to
            use to process multiple chromosomes in parallel. Pass -1 to use as
            many threads as there are CPUs. Pass 0 to process the chromosomes
            serially.
        verbose : bool
            Pass False to silence reporting of progress to stderr.
        """
        if n_bins == -1:
            n_bins = self.dist_thresh_max / 5

        if chrom is None:
            if n_threads:
                parallel_apply(
                    self.prepare_data,
                    [{'chrom': c, 'norm': norm, 'n_bins': n_bins,
                      'verbose': False}
                     for c in self.chroms],
                    n_threads=n_threads
                )
            else:
                for chrom in self.chroms:
                    self.prepare_data(chrom=chrom, norm=norm, n_bins=n_bins)
            return
        eprint('preparing data for chrom %s' % chrom)
        eprint('  loading bias', skip=not verbose)
        bias = self.load_bias(chrom)

        eprint('  computing union pixel set', skip=not verbose)
        row, col = sparse_union(
            [pattern.replace('<chrom>', chrom)
             for pattern in self.raw_npz_patterns],
            dist_thresh=self.dist_thresh_max,
            bias=bias
        )

        eprint('  loading raw data', skip=not verbose)
        raw = np.zeros((len(row), len(self.raw_npz_patterns)), dtype=int)
        for i, pattern in enumerate(self.raw_npz_patterns):
            raw[:, i] = sparse.load_npz(pattern.replace('<chrom>', chrom)) \
                .tocsr()[row, col]

        eprint('  loading balanced data', skip=not verbose)
        balanced = np.zeros((len(row), len(self.raw_npz_patterns)), dtype=float)
        for r, pattern in enumerate(self.raw_npz_patterns):
            balanced[:, r] = sparse.load_npz(pattern.replace('<chrom>', chrom))\
                .tocsr()[row, col] / (bias[row, r] * bias[col, r])

        eprint('  computing size factors', skip=not verbose)
        if 'conditional' in norm:
            size_factors = scaling.__dict__[norm](balanced, col - row,
                                                  n_bins=n_bins)
        else:
            size_factors = scaling.__dict__[norm](balanced)
        scaled = balanced / size_factors

        eprint('  computing disp_idx', skip=not verbose)
        dist = col - row
        mean = np.dot(scaled, self.design) / np.sum(self.design, axis=0).values
        disp_idx = np.all(mean >= self.mean_thresh, axis=1) & \
            (dist >= self.dist_thresh_min)

        if self.loop_patterns:
            eprint('  making loop_idx', skip=not verbose)
            loop_pixels = set().union(
                *sum((load_clusters(pattern.replace('<chrom>', chrom))
                      for pattern in self.loop_patterns.values()), []))
            loop_idx = np.array([True if pixel in loop_pixels else False
                                 for pixel in zip(row[disp_idx],
                                                  col[disp_idx])])
            self.save_data(loop_idx, 'loop_idx', chrom)

        eprint('  saving data to disk', skip=not verbose)
        self.save_data(row, 'row', chrom)
        self.save_data(col, 'col', chrom)
        self.save_data(raw, 'raw', chrom)
        self.save_data(size_factors, 'size_factors', chrom)
        self.save_data(scaled, 'scaled', chrom)
        self.save_data(disp_idx, 'disp_idx', chrom)

    def estimate_disp(self, estimator='qcml', frac=None, auto_frac_factor=15.,
                      weighted_lowess=True, n_threads=-1):
        """
        Estimates dispersion parameters.

        Parameters
        ----------
        estimator : 'cml', 'qcml', 'mme', or a function
            Pass 'cml', 'qcml', 'mme' to use conditional maximum likelihood
            (CML), quantile-adjusted CML (qCML), or method of moments estimation
            (MME) to estimate the dispersion within each bin. Pass a function
            that takes in a (pixels, replicates) shaped array of data and
            returns a dispersion value to use that instead.
        frac : float, optional
            The lowess smoothing fraction to use when fitting the distance vs
            dispersion trend. Pass None to choose a value automatically.
        auto_frac_factor : float
            When ``frac`` is None, this factor scales the automatically
            determined fraction parameter.
        weighted_lowess : bool
            Whether or not to use a weighted lowess fit when fitting the
            smoothed dispersion curve.
        n_threads : int
            The number of threads (technically GIL-avoiding child processes) to
            use to process multiple distance scales in parallel. Pass -1 to use
            as many threads as there are CPUs. Pass 0 to process the distance
            scales serially.
        """
        eprint('estimating dispersion')
        estimator = dispersion.__dict__[estimator] \
            if estimator in dispersion.__dict__ else estimator
        lowess_fn = weighted_lowess_fit if weighted_lowess else lowess_fit

        eprint('  loading data')
        disp_idx, disp_idx_offsets = self.load_data('disp_idx', 'all')
        row, offsets = self.load_data('row', 'all', idx=disp_idx)
        col, _ = self.load_data('col', 'all', idx=disp_idx)
        raw, _ = self.load_data('raw', 'all', idx=disp_idx)
        dist = col - row
        f = np.ones_like(raw, dtype=float)
        for i, chrom in enumerate(self.chroms):
            chrom_slice = slice(offsets[i], offsets[i+1])
            row_chrom = row[chrom_slice]
            col_chrom = col[chrom_slice]
            disp_idx_chrom = disp_idx[disp_idx_offsets[i]:disp_idx_offsets[i+1]]
            bias = self.load_bias(chrom)
            size_factors = self.load_data('size_factors', chrom)[disp_idx_chrom]
            f[chrom_slice] = bias[row_chrom, :] * bias[col_chrom, :] \
                * size_factors

        def _estimate(raw_slice, f_slice):
            return np.nan if not raw_slice.size \
                else estimator(raw_slice, f=f_slice)

        disp_per_dist = np.zeros((self.dist_thresh_max+1, self.design.shape[1]))
        disp = np.zeros((disp_idx.sum(), self.design.shape[1]))
        for c, cond in enumerate(self.design.columns):
            eprint('  estimating dispersion for condition %s' % cond)
            if n_threads:
                disp_per_dist[:, c] = parallel_map(
                    _estimate,
                    [{'raw_slice': raw[dist == d, :][:, self.design[cond]],
                      'f_slice': f[dist == d, :][:, self.design[cond]]}
                     for d in range(self.dist_thresh_max+1)],
                    n_threads=n_threads
                )
            else:
                for d in tqdm(range(self.dist_thresh_max+1)):
                    raw_slice = raw[dist == d, :][:, self.design[cond]]
                    f_slice = f[dist == d, :][:, self.design[cond]]
                    disp_per_dist[d, c] = np.nan if not raw_slice.size \
                        else estimator(raw_slice, f=f_slice)

            eprint('  fitting distance vs dispersion relationship')
            idx = np.isfinite(disp_per_dist[:, c])
            x = np.arange(self.dist_thresh_max+1)[idx]
            y = disp_per_dist[:, c][idx]
            lowess_kwargs = {'left_boundary': y[0]}
            if frac is not None:
                lowess_kwargs['frac'] = frac
            if weighted_lowess:
                lowess_kwargs['auto_frac_factor'] = auto_frac_factor
            disp_fn = lowess_fn(x, y, **lowess_kwargs)
            disp[:, c] = disp_fn(dist)
            self.save_disp_fn(cond, disp_fn)

        eprint('  saving estimated dispersions to disk')
        self.save_data(disp, 'disp', offsets)
        self.save_data(disp_per_dist, 'disp_per_dist')

    def lrt(self, chrom=None, refit_mu=True, n_threads=-1, verbose=True):
        """
        Runs the likelihood ratio test to test for differential interactions.

        Parameters
        ----------
        chrom : str
            The name of the chromosome to run the LRT for. Pass None to run for
            all chromosomes in series.
        refit_mu : bool
            Pass True to refit the mean parameters in the NB models being
            compared in the LRT. Pass False to use the means across replicates
            directly, which is simpler and slightly faster but technically
            violates the assumptions of the LRT.
        n_threads : int
            The number of threads (technically GIL-avoiding child processes) to
            use to process multiple chromosomes in parallel. Pass -1 to use as
            many threads as there are CPUs. Pass 0 to process the chromosomes
            serially.
        verbose : bool
            Pass False to silence reporting of progress to stderr.
        """
        if chrom is None:
            if n_threads:
                parallel_apply(
                    self.lrt,
                    [{'chrom': c, 'refit_mu': refit_mu, 'verbose': False}
                     for c in self.chroms],
                    n_threads=n_threads
                )
            else:
                for chrom in self.chroms:
                    self.lrt(chrom=chrom, refit_mu=refit_mu)
            return
        eprint('running LRT for chrom %s' % chrom)
        eprint('  loading data', skip=not verbose)
        bias = self.load_bias(chrom)
        # we load size_factors the slow way because we don't know in advance if
        # it's 2 dimensional or not
        size_factors = self.load_data('size_factors', chrom)
        disp_idx = self.load_data('disp_idx', chrom)
        row = self.load_data('row', chrom, idx=disp_idx)
        col = self.load_data('col', chrom, idx=disp_idx)
        raw = self.load_data('raw', chrom, idx=disp_idx)
        disp = self.load_data('disp', chrom)

        eprint('  computing LRT results', skip=not verbose)
        if len(size_factors.shape) == 2:
            f = bias[row] * bias[col] * size_factors[disp_idx, :]
        else:
            f = bias[row] * bias[col] * size_factors
        pvalues, llr, mu_hat_null, mu_hat_alt = lrt(
            raw, f, np.dot(disp, self.design.values.T),
            self.design.values, refit_mu=refit_mu)

        eprint('  saving results to disk', skip=not verbose)
        self.save_data(pvalues, 'pvalues', chrom)
        self.save_data(llr, 'llr', chrom)
        self.save_data(mu_hat_null, 'mu_hat_null', chrom)
        self.save_data(mu_hat_alt, 'mu_hat_alt', chrom)

    def bh(self):
        """
        Applies BH-FDR control to p-values across all chromosomes to obtain
        q-values.

        Should only be run after all chromosomes have been processed through
        p-values.
        """
        eprint('applying BH-FDR correction')
        if self.loop_patterns:
            loop_idx, _ = self.load_data('loop_idx', 'all')
        else:
            loop_idx = None
        pvalues, offsets = self.load_data('pvalues', 'all', idx=loop_idx)
        all_qvalues = adjust_pvalues(pvalues)
        for i, chrom in enumerate(self.chroms):
            self.save_data(all_qvalues[offsets[i]:offsets[i+1]], 'qvalues',
                           chrom)

    def run_to_qvalues(self, norm='conditional_mor', n_bins_norm=-1,
                       estimator='qcml', frac=None, auto_frac_factor=15.,
                       weighted_lowess=True, refit_mu=True, n_threads=-1,
                       verbose=True):
        """
        Shortcut method to run the analysis to q-values.

        Parameters
        ----------
        norm : str
            The method to use to account for differences in sequencing depth.
            Valid options are:
                * simple_scaling: scale each replicate to equal total depth
                * median_of_ratios: use median of ratios normalization, ignoring
                  pixels at which any replicate has a zero
                * conditional_scaling: apply simple scaling independently at
                  each distance scale
                * conditional_mor: apply median of ratios independently at each
                  distance scale
        n_bins_norm : int, optional
            Number of distance bins to use during scaling normalization if
            ``norm`` is one of the conditional options. Pass 0 or None to match
            pixels by exact distance. Pass -1 to use a reasonable default: 1/5
            of ``self.dist_thesh_max``.
        estimator : 'cml', 'qcml', 'mme', or a function
            Pass 'cml', 'qcml', 'mme' to use conditional maximum likelihood
            (CML), qnorm-CML (qCML), or method of moments estimation (MME) to
            estimate the dispersion within each bin. Pass a function that takes
            in a (pixels, replicates) shaped array of data and returns a
            dispersion value to use that instead.
        frac : float, optional
            The lowess smoothing fraction to use when fitting the distance vs
            dispersion trend. Pass None to choose a value automatically.
        auto_frac_factor : float
            When ``frac`` is None, this factor scales the automatically
            determined fraction parameter.
        weighted_lowess : bool
            Whether or not to use a weighted lowess fit when fitting the
            smoothed dispersion curve.
        refit_mu : bool
            Pass True to refit the mean parameters in the NB models being
            compared in the LRT. Pass False to use the means across replicates
            directly, which is simpler and slightly faster but technically
            violates the assumptions of the LRT.
        n_threads : int
            The number of threads (technically GIL-avoiding child processes) to
            use to process multiple chromosomes in parallel. Pass -1 to use as
            many threads as there are CPUs. Pass 0 to process the chromosomes
            serially.
        verbose : bool
            Pass False to silence reporting of progress to stderr.
        """
        self.prepare_data(norm=norm, n_bins=n_bins_norm, n_threads=n_threads)
        self.estimate_disp(
            estimator=estimator, frac=frac, auto_frac_factor=auto_frac_factor,
            weighted_lowess=weighted_lowess, n_threads=n_threads)
        self.lrt(refit_mu=refit_mu, n_threads=n_threads)
        self.bh()

    def threshold(self, chrom=None, fdr=0.05, cluster_size=3, n_threads=-1):
        """
        Thresholds and clusters significantly differential pixels.

        Should only be run after q-values have been obtained.

        Parameters
        ----------
        chrom : str
            The name of the chromosome to threshold. Pass None to threshold all
            chromosomes in series.
        fdr : float or list of float
            The FDR to threshold on. Pass a list to do a sweep in series.
        cluster_size : int or list of int
            Clusters smaller than this size will be filtered out. Pass a list to
            do a sweep in series.,
        n_threads : int
            The number of threads (technically GIL-avoiding child processes) to
            use to process multiple chromosomes in parallel. Pass -1 to use as
            many threads as there are CPUs. Pass 0 to process the chromosomes
            serially.
        """
        if chrom is None:
            if n_threads:
                parallel_apply(
                    self.threshold,
                    [{'chrom': c, 'fdr': fdr, 'cluster_size': cluster_size}
                     for c in self.chroms],
                    n_threads=n_threads
                )
            else:
                for chrom in self.chroms:
                    self.threshold(chrom=chrom, fdr=fdr,
                                   cluster_size=cluster_size)
            return
        eprint('thresholding and clustering chrom %s' % chrom)
        # load everything
        disp_idx = self.load_data('disp_idx', chrom)
        loop_idx = self.load_data('loop_idx', chrom) \
            if self.loop_patterns else np.ones(disp_idx.sum(), dtype=bool)
        row = self.load_data('row', chrom, idx=(disp_idx, loop_idx))
        col = self.load_data('col', chrom, idx=(disp_idx, loop_idx))
        qvalues = self.load_data('qvalues', chrom)

        # upgrade fdr and cluster_size to list
        if not hasattr(fdr, '__len__'):
            fdr = [fdr]
        if not hasattr(cluster_size, '__len__'):
            cluster_size = [cluster_size]

        for f in fdr:
            sig_clusters, insig_clusters = threshold_and_cluster(
                qvalues, row, col, fdr)
            for s in cluster_size:
                # threshold on cluster size and save to disk
                save_clusters(size_filter(sig_clusters, s),
                              '%s/sig_%g_%i_%s.json' %
                              (self.outdir, f, s, chrom))
                save_clusters(size_filter(insig_clusters, s),
                              '%s/insig_%g_%i_%s.json' %
                              (self.outdir, f, s, chrom))

    def classify(self, chrom=None, fdr=0.05, cluster_size=3, n_threads=-1):
        """
        Classifies significantly differential pixels according to which
        condition they are strongest in.

        Parameters
        ----------
        chrom : str
            The chromosome to classify significantly differential pixels on.
            Pass None to run for all chromosomes in series.
        fdr : float or list of float
            The FDR threshold used to identify clusters of significantly
            differential pixels via ``self.threshold()``. Pass a list to do a
            sweep in series.
        cluster_size : int or list of int
            The cluster size threshold used to identify clusters of
            significantly differential pixels via ``threshold_chrom()``. Pass a
            list to do a sweep in series.
        n_threads : int
            The number of threads (technically GIL-avoiding child processes) to
            use to process multiple chromosomes in parallel. Pass -1 to use as
            many threads as there are CPUs. Pass 0 to process the chromosomes
            serially.
        """
        if chrom is None:
            if n_threads:
                parallel_apply(
                    self.classify,
                    [{'chrom': c, 'fdr': fdr, 'cluster_size': cluster_size}
                     for c in self.chroms],
                    n_threads=n_threads
                )
            else:
                for chrom in self.chroms:
                    self.classify(chrom=chrom, fdr=fdr,
                                  cluster_size=cluster_size)
            return
        eprint('classifying differential interactions on chrom %s' % chrom)
        # load everything
        disp_idx = self.load_data('disp_idx', chrom)
        loop_idx = self.load_data('loop_idx', chrom) \
            if self.loop_patterns else np.ones(disp_idx.sum(), dtype=bool)
        row = self.load_data('row', chrom, idx=(disp_idx, loop_idx))
        col = self.load_data('col', chrom, idx=(disp_idx, loop_idx))
        mu_hat_alt = self.load_data('mu_hat_alt', chrom, idx=loop_idx)

        # upgrade fdr and cluster_size to list
        if not hasattr(fdr, '__len__'):
            fdr = [fdr]
        if not hasattr(cluster_size, '__len__'):
            cluster_size = [cluster_size]

        for f in fdr:
            for s in cluster_size:
                infile = '%s/sig_%g_%i_%s.json' % (self.outdir, f, s, chrom)
                if not os.path.isfile(infile):
                    self.threshold(chrom=chrom, fdr=f, cluster_size=s)
                sig_clusters = load_clusters(infile)
                class_clusters = classify(row, col, mu_hat_alt, sig_clusters)
                for i, c in enumerate(class_clusters):
                    save_clusters(
                        c, '%s/%s_%g_%i_%s.json' %
                           (self.outdir, self.design.columns[i], f, s, chrom))



import numpy as np
import scipy.sparse as sparse
from scipy.ndimage import zoom


def deconvolute(matrix, bias, invert=False):
    """
    Applies bias factors to a sparse matrix.

    Parameters
    ----------
    matrix : scipy.sparse.spmatrix
        The matrix to bias. Will be converted to CSR by this function.
    bias : np.ndarray
        The dense bias vector.
    invert : bool
        Whether or not to invert the bias before applying it. By default this
        function multiplies the matrix by the bias. Infinite bias (1 / 0) will
        be converted to zero bias to allow rows with infinite bias to be dropped
        from the resulting sparse matrix.

    Returns
    -------
    scipy.sparse.csr_matrix
        The deconvoluted matrix.
    """
    csr = matrix.tocsr()
    if invert:
        inf_idx = bias == 0
        bias[inf_idx] = 1
        bias = 1 / bias
        bias[inf_idx] = 0
    bias_csr = sparse.diags([bias], [0])
    biased_csr = bias_csr.dot(csr)
    biased_csr = biased_csr.dot(bias_csr)
    return biased_csr


def wipe_distances(matrix, min_dist, max_dist):
    """
    Eliminates entries from a sparse matrix outside of a specified distance
    range.

    Parameters
    ----------
    matrix : scipy.sparse.spmatrix
        The matrix to wipe. Will be converted to COO by this function.
    min_dist, max_dist : int
        The minimum and maximum distance allowed, respectively, in bin units.

    Returns
    -------
    scipy.sparse.coo_matrix
        The wiped matrix.
    """
    coo = matrix.tocoo()
    dist = coo.col - coo.row
    coo.data[(dist < min_dist) | (dist > max_dist)] = 0
    coo.eliminate_zeros()
    return coo


def sparse_intersection(fnames, bias=None):
    """
    Computes the intersection set of (row, col) pairs across multiple sparse
    matrices.

    Parameters
    ----------
    fnames : list of str
        File paths to sparse matrices loadable by ``scipy.sparse.load_npz()``.
        Will be converted to COO by this function.
    bias : np.ndarray, optional
        Pass the bias matrix to drop rows and columns with bias factors of
        zero in any replicate.

    Returns
    -------
    row, col : np.ndarray
        The intersection set of (row, col) pairs.
    """
    csr_sum_coo = reduce(
        lambda x, y: x + y,
        ((deconvolute(sparse.load_npz(fname), bias[:, i]) > 0).astype(int)
         for i, fname in enumerate(fnames))).tocoo()
    full_idx = csr_sum_coo.data == len(fnames)
    return csr_sum_coo.row[full_idx], csr_sum_coo.col[full_idx]


def sparse_union(fnames, dist_thresh=1000, bias=None, size_factors=None,
                 mean_thresh=0.0):
    """
    Computes the intersection set of (row, col) pairs across multiple sparse
    matrices.

    Parameters
    ----------
    fnames : list of str
        File paths to sparse matrices loadable by ``scipy.sparse.load_npz()``.
        Will be converted to COO by this function.
    dist_thresh : int
        The maximum distance allowed, respectively, in bin units.
    bias : np.ndarray, optional
        Rectangular matrix containing bias factors for each bin (rows) and each
        replicate (columns).
    size_factors : np.ndarray, optional
        Size factors for each replicate.
    mean_thresh : float
        Minimum mean value (in normalized space) to keep pixels for.

    Returns
    -------
    row, col : np.ndarray
        The union set of (row, col) pairs.
    """
    if size_factors is None:
        size_factors = np.ones(len(fnames))
    csr_sum_coo = reduce(
        lambda x, y: x + y,
        (wipe_distances((deconvolute(sparse.load_npz(fname), bias[:, i],
                                     invert=True)
                         if bias is not None else sparse.load_npz(fname))
                        / size_factors[i], 0, dist_thresh)
         for i, fname in enumerate(fnames))).tocoo()
    full_idx = (csr_sum_coo.data >= len(fnames)*mean_thresh) \
        & np.isfinite(csr_sum_coo.data)
    return csr_sum_coo.row[full_idx], csr_sum_coo.col[full_idx]


def select_matrix(row_slice, col_slice, row, col, data, symmetrize=True):
    """
    Slices out a dense matrix from COO-formatted sparse data, filling empty
    values with ``np.nan``.

    Parameters
    ----------
    row_slice, col_slice : slice
        Row and column slice to use, respectively.
    row, col, data : np.ndarray
        COO-format sparse matrix to be sliced.
    symmetrize : bool
        Pass True to fill in lower-triangle points of the matrix.

    Returns
    -------
    np.ndarray
        Dense matrix.
    """
    r_start, r_stop = row_slice.start, row_slice.stop
    c_start, c_stop = col_slice.start, col_slice.stop
    idx = (row >= r_start) & (row < r_stop) & (col >= c_start) & (col < c_stop)
    matrix = np.ones((r_stop - r_start, c_stop - c_start)) * np.nan
    matrix[row[idx]-r_start, col[idx]-c_start] = data[idx]
    if symmetrize:
        t_idx = (col >= r_start) & (col < r_stop) & (row >= c_start) & \
            (row < c_stop)
        matrix[col[t_idx]-r_start, row[t_idx]-c_start] = data[t_idx]
    return matrix


def dilate(matrix, doublings):
    """
    Doubles the "resolution" of a matrix using nearest-neighbor interpolation.

    Parameters
    ----------
    matrix : np.ndarray
        Input matrix.
    doublings : int
        The number of times the resolution will be doubled.

    Returns
    -------
    np.ndarray
        The dilated matrix.
    """
    for _ in range(doublings):
        matrix = zoom(matrix, 2, order=0)
    return matrix



import matplotlib.pyplot as plt

from lib5c.util.plotting import plotter

from hic3defdr.util.matrices import select_matrix


@plotter
def plot_heatmap(row, col, data, row_slice, col_slice, cmap='Reds', vmin=0,
                 vmax=100, despine=False, **kwargs):
    """
    Plots a simple heatmap of a slice of a contact matrix.

    Parameters
    ----------
    row, col, data : np.ndarray
        Sparse COO-style description of the contact matrix.
    row_slice, col_slice : slice
        The row and column slice, respectively, to plot.
    cmap : matplotlib colormap
        The colormap to use for the heatmap.
    vmin, vmax : float
        The vmin and vmax to use for the heatmap colorscale.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis
        The axis plotted on.
    """
    matrix = select_matrix(row_slice, col_slice, row, col, data)
    plt.imshow(matrix, cmap=cmap, vmin=vmin, vmax=vmax, interpolation='none')
    plt.xticks([-0.5, matrix.shape[1]-0.5], [row_slice.start, row_slice.stop])
    plt.yticks([-0.5, matrix.shape[0]-0.5], [col_slice.start, col_slice.stop])



def context():
    """
    Infers the context in which the current program is being executed.

    https://stackoverflow.com/a/47428575

    Returns
    -------
    str
        The context, either 'colab', 'jupyter', 'ipython', or 'terminal'.
    """
    try:
        ipy_str = str(type(get_ipython()))
        if 'colab' in ipy_str:
            return 'colab'
        if 'zmqshell' in ipy_str:
            return 'jupyter'
        if 'terminal' in ipy_str:
            return 'ipython'
    except NameError:
        return 'terminal'


try:
    if context() in ['jupyter', 'colab']:
        from tqdm import tqdm_notebook as tqdm
    else:
        from tqdm import tqdm
    tqdm_avail = True
    import tqdm as tqdm_module
except ImportError:
    tqdm_avail = False
    tqdm_module = None
    tqdm = None


def tqdm_maybe(iter, **kwargs):
    """
    Drop-in replacement from ``tqdm.tqdm()`` except will simply do nothing if
    ``tqdm`` is not present and will use ``tqdm.tqdm_notebook()`` if run in a
    notebook.

    Parameters
    ----------
    iter : iterable
        The iterable to wrap.
    kwargs : kwargs
        Will be passed through to ``tqdm.tqdm()``.

    Returns
    -------
    iterator
        Wrapped by ``tqdm`` if it was installed, or just ``iter`` otherwise.
    """
    if tqdm_avail:
        return tqdm(iter, **kwargs)
    return iter



from setuptools import setup, find_packages

import versioneer

readme_note = """\
.. note::
   For the latest source, discussion, etc, please visit the
   `Bitbucket repository <https://bitbucket.org/creminslab/hic3defdr>`_\n\n
"""

with open('README.md') as fobj:
    long_description = readme_note  # + fobj.read()

extras_require = {
    'evaluation': ['scikit-learn>=0.20.3'],
    'test': ['nose>=1.3.7', 'nose-exclude>=0.5.0', 'flake8>=3.4.1'],
    'progress': ['tqdm>=4.32.2'],
}
extras_require['complete'] = sorted(set(sum(extras_require.values(), [])))

setup(
    name='hic3defdr',
    version='0.0.9',
    #cmdclass=versioneer.get_cmdclass(),
    description='a genome-scale differential loop finder',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Thomas Gilgenast',
    url='https://bitbucket.org/creminslab/hic3defdr',
    packages=find_packages(),
    install_requires=[
        'numpy>=1.14.0',
        'scipy>=1.2.0',
        'matplotlib>=2.1.1',
        'mpl_scatter_density>=0.6',
        'seaborn>=0.8.0',
        'pandas>=0.21.0',
        'lib5c>=0.5.3',
        'dill>=0.2.9',
    ],
    extras_require=extras_require,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
)



import numpy as np
import dill as pickle


class CoreHiC3DeFDR(object):
    """
    Mixin class providing core saving and loading functionality for HiC3DeFDR.
    """
    @property
    def picklefile(self):
        return '%s/pickle' % self.outdir

    @classmethod
    def load(cls, outdir):
        """
        Loads a HiC3DeFDR analysis object from disk.

        It is safe to have multiple instances of the same analysis open at once.

        Parameters
        ----------
        outdir : str
            Folder path to where the HiC3DeFDR was saved.

        Returns
        -------
        HiC3DeFDR
            The loaded object.
        """
        with open('%s/pickle' % outdir, 'rb') as handle:
            return cls(outdir=outdir, **pickle.load(handle))

    def load_bias(self, chrom):
        """
        Loads the bias matrix for one chromosome.

        The rows of the bias matrix correspond to bin indices along the
        chromosome. The columns correspond to the replicates.

        The bias factors for bins that fail ``bias_thresh`` are set to zero.
        This is designed so that all pixels in these bins get dropped during
        union pixel set computation.

        Parameters
        ----------
        chrom : str
            The name of the chromosome to load the bias matrix for.

        Returns
        -------
        np.ndarray
            The bias matrix.
        """
        bias = np.array([np.loadtxt(pattern.replace('<chrom>', chrom))
                         for pattern in self.bias_patterns]).T
        bias[(np.any(bias < self.bias_thresh, axis=1)) |
             (np.any(bias > 1. / self.bias_thresh, axis=1)), :] = 0
        return bias

    def load_data(self, name, chrom=None, idx=None):
        """
        Loads arbitrary data for one chromosome or all chromosomes.

        Parameters
        ----------
        name : str
            The name of the data to load.
        chrom : str, optional
            The name of the chromosome to load data for. Pass None if this data
            is not organized by chromosome. Pass 'all' to load data for all
            chromosomes.
        idx : np.ndarray or tuple of np.ndarray, optional
            Pass a boolean array to load only a subset of the data. Pass a tuple
            of exactly two boolean arrays to chain the indices.

        Returns
        -------
        data or (concatenated_data, offsets) : np.ndarray
            The loaded data for one chromosome, or a tuple of the concatenated
            data and an array of offsets per chromosome. ``offsets`` satisfies
            the following properties: ``offsets[0] == 0``,
            ``offsets[-1] == concatenated_data.shape[0]``, and
            ``concatenated_data[offsets[i]:offsets[i+1], :]`` contains the data
            for the ``i``th chromosome.
        """
        # index chaining
        if type(idx) == tuple:
            big_idx, small_idx = idx
            big_idx = big_idx.copy()
            big_idx[np.where(big_idx)[0][~small_idx]] = False
            idx = big_idx

        # tackle simple cases first
        if chrom is None:
            fname = '%s/%s.npy' % (self.outdir, name)
        elif chrom != 'all':
            fname = '%s/%s_%s.npy' % (self.outdir, name, chrom)
        else:
            fname = None
        if fname is not None:
            if idx is None:
                return np.load(fname)
            else:
                return np.load(fname, mmap_mode='r')[idx]

        # idx is genome-wide, this tracks where we are in idx so that we can
        # find a subset of idx that aligns with the current chrom
        idx_offset = 0

        # list of data arrays per chromosome
        all_data = []

        # running total of the sizes of the elements of all data
        offset = 0

        # saves value of offset after each chrom
        offsets = [0]

        # loop over chroms
        for chrom in self.chroms:
            fname = '%s/%s_%s.npy' % (self.outdir, name, chrom)
            if idx is not None:
                data = np.load(fname, mmap_mode='r')
                full_data_size = data.shape[0]
                data = data[idx[idx_offset:idx_offset+full_data_size]]
                idx_offset += full_data_size
            else:
                data = np.load(fname)
            offset += data.shape[0]
            offsets.append(offset)
            all_data.append(data)
        return np.concatenate(all_data), np.array(offsets)

    def save_data(self, data, name, chrom=None):
        """
        Saves arbitrary data for one chromosome to disk.

        Parameters
        ----------
        data : np.ndarray
            The data to save.
        name : str
            The name of the data to save.
        chrom : str or np.ndarray, optional
            The name of the chromosome to save data for, or None if this data is
            not organized by chromosome. Pass an np.ndarray of offsets to save
            data for all chromosomes.
        """
        if chrom is None:
            np.save('%s/%s.npy' % (self.outdir, name), data)
        elif isinstance(chrom, np.ndarray):
            for i, c in enumerate(self.chroms):
                self.save_data(data[chrom[i]:chrom[i + 1]], name, c)
        else:
            np.save('%s/%s_%s.npy' % (self.outdir, name, chrom), data)

    def load_disp_fn(self, cond):
        """
        Loads the fitted dispersion function for a specific condition from disk.

        Parameters
        ----------
        cond : str
            The condition to load the dispersion function for.

        Returns
        -------
        function
            Vectorized. Takes in the value of the covariate the dispersion was
            fitted against and returns the appropriate dispersion.
        """
        picklefile = '%s/disp_fn_%s.pickle' % (self.outdir, cond)
        with open(picklefile, 'rb') as handle:
            return pickle.load(handle)

    def save_disp_fn(self, cond, disp_fn):
        """
        Saves the fitted dispersion function for a specific condition and
        chromosome to disk.

        Parameters
        ----------
        cond : str
            The condition to save the dispersion function for.
        disp_fn : function
            The dispersion function to save.
        """
        picklefile = '%s/disp_fn_%s.pickle' % (self.outdir, cond)
        with open(picklefile, 'wb') as handle:
            return pickle.dump(disp_fn, handle, -1)
